# SPDX-FileCopyrightText: 2023 Anthony Zimmermann
#
# SPDX-License-Identifier: GPL-2.0-only

import importlib.resources
import math

from PySide6.QtCore import QPoint
from PySide6.QtCore import QRect
from PySide6.QtCore import QRectF
from PySide6.QtCore import Qt
from PySide6.QtGui import QFontMetricsF
from PySide6.QtGui import QPainter
from PySide6.QtGui import QPalette
from PySide6.QtGui import QTextLayout
from PySide6.QtGui import QTextOption
from PySide6.QtWidgets import QApplication
from PySide6.QtWidgets import QCommonStyle
from PySide6.QtWidgets import QStyle
from PySide6.QtWidgets import QStyledItemDelegate
from PySide6.QtWidgets import QStyleOptionViewItem

from photoassistant.gui.qt.utils.qtutils import QPixmapLoader


class PhotoViewItemDelegate(QStyledItemDelegate):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        app = QApplication.instance()
        assert isinstance(app, QApplication)

        self.photo_view_text_height = math.ceil(2 * QFontMetricsF(app.font()).lineSpacing())

        screen_dots_per_mm = app.primaryScreen().physicalDotsPerInch() / 25.4
        self.image_margin = int(1.2 * screen_dots_per_mm)

        with importlib.resources.path("photoassistant.gui.resources", "icon-loading.png") as icon_path:
            self.placeholder_pixmap = QPixmapLoader.load(str(icon_path.resolve()))

    def paint(self, painter, option, index):
        return self.paint_with_pixmap(painter, option, index)

    def paint_with_pixmap(self, painter, option, index, pixmap=None):
        painter.save()
        painter.setRenderHints(QPainter.RenderHints.Antialiasing)

        # init paint option, widget and style
        # this creates an OptionStyleOptionViewItem (opt) as is done in the
        # original function QStyledItemDelegate.paint such that we can reuse
        # some code:
        # - setClipRegion(...)
        # The entire widget in the original function would be drawn with
        # 'style.drawControl(QStyle.ControlElement.CE_ItemViewItem, opt, ...)'.
        opt = QStyleOptionViewItem(option)
        self.initStyleOption(opt, index)
        widget = option.widget
        style = widget.style()

        # The model data of different roles of 'index' is transferred into 'opt'
        # as following:
        # (Extracted from source code documentation at
        # src/widgets/intemviews/qstyleditemdelegate.cpp):
        # option                        | Role                  | Accepted Types
        # ------------------------------+-----------------------+---------------
        # opt->backgroundBrush          | Qt::BackgroundRole    | QBrush
        # opt->features | HasCheckInd...| Qt::CheckStateRole    | Qt::CheckState
        # opt->checkState               |                       |
        # opt->features | HasDecoration | Qt::DecorationRole    | QIcon,QPixmap,
        # opt->icon                     |                       | QImage and
        # opt->decorationSize           |                       | QColor
        # opt->features | HasDisplay    | Qt::DisplayRole       | QString and
        # opt->text                     |                       | types with a
        #                               |                       | string repr
        # opt->font                     | Qt::FontRole          | QFont
        # opt->fontMetrics              |                       |
        # opt->displayAlignment         | Qt::TextAlignmentRole | Qt::Alignment
        # opt->palette.brush[:,Text]    | Qt::ForegroundRole    | QBrush

        # intersect clipping regions of painter and option
        if painter.hasClipping():
            painter.setClipRegion(painter.clipRegion().intersected(opt.rect))
        else:
            painter.setClipRegion(opt.rect)

        image_rect = option.rect.adjusted(
            self.image_margin,
            self.image_margin,
            -self.image_margin,
            -(self.image_margin + self.photo_view_text_height),
        )
        text_rect = QRect(
            image_rect.x(),
            (image_rect.bottom() + (self.image_margin >> 1)),
            image_rect.width(),
            self.photo_view_text_height,
        )
        selected_highlight_rect = option.rect if opt.showDecorationSelected else text_rect

        color_group = QPalette.ColorGroup.Disabled
        if QStyle.StateFlag.State_Enabled in opt.state:
            if QStyle.StateFlag.State_Active not in opt.state:
                color_group = QPalette.ColorGroup.Inactive
            else:
                color_group = QPalette.ColorGroup.Normal

        # draw background
        if QStyle.StateFlag.State_Selected in opt.state:
            painter.fillRect(selected_highlight_rect, opt.palette.brush(color_group, QPalette.ColorRole.Highlight))

        # draw image
        photo_object = index.data(Qt.ItemDataRole.UserRole)
        if photo_object:
            if pixmap is None:
                pixmap = self.placeholder_pixmap

            if pixmap:
                pixmap = pixmap.scaled(
                    image_rect.width(),
                    image_rect.height(),
                    Qt.AspectRatioMode.KeepAspectRatio,
                )
                pixmap_rect = QRectF(
                    image_rect.x() + ((image_rect.width() - pixmap.width()) / 2),
                    image_rect.y() + ((image_rect.height() - pixmap.height()) / 2),
                    pixmap.width(),
                    pixmap.height(),
                ).toRect()
                painter.drawPixmap(pixmap_rect, pixmap)

        # draw filename of image (text)
        if QStyleOptionViewItem.ViewItemFeature.HasDisplay in opt.features and opt.text:
            if QStyle.StateFlag.State_Selected in opt.state:
                painter.setPen(opt.palette.color(color_group, QPalette.ColorRole.HighlightedText))
            else:
                painter.setPen(opt.palette.color(color_group, QPalette.ColorRole.Text))

            if QStyle.StateFlag.State_Editing in opt.state:
                painter.setPen(opt.palette.color(color_group, QPalette.ColorRole.Text))
                painter.drawRect(text_rect.adjusted(0, 0, -1, -1))

            text_margin = style.pixelMetric(QCommonStyle.PixelMetric.PM_FocusFrameHMargin, None, widget)
            text_rect = text_rect.adjusted(text_margin, 0, -text_margin, 0)

            text_option = QTextOption()
            text_option.setWrapMode(QTextOption.WrapMode.WrapAnywhere)
            text_option.setTextDirection(opt.direction)
            text_option.setAlignment(QStyle.visualAlignment(opt.direction, opt.displayAlignment))

            path_text = str(opt.text)
            text_layout = QTextLayout(path_text, opt.font)
            text_layout.setTextOption(text_option)

            line_width = text_rect.width()
            line_height = opt.fontMetrics.lineSpacing()
            text_draw_pos = text_rect.topLeft()

            text_layout.beginLayout()
            while True:
                line = text_layout.createLine()
                if not line.isValid():
                    break
                line.setLineWidth(line_width)
                if (text_draw_pos.y() + line_height) < text_rect.bottom():
                    line.draw(painter, text_draw_pos)
                    text_draw_pos += QPoint(0, line_height)
                else:
                    elided_last_line_path_text = opt.fontMetrics.elidedText(
                        path_text[line.textStart():],
                        Qt.TextElideMode.ElideRight,
                        text_rect.width(),
                    )
                    text_draw_pos += QPoint(0, opt.fontMetrics.ascent())
                    painter.drawText(
                        text_draw_pos,
                        elided_last_line_path_text,
                    )
                    break
            text_layout.endLayout()

        painter.restore()



# SPDX-FileCopyrightText: 2023 Anthony Zimmermann
#
# SPDX-License-Identifier: GPL-2.0-only

import datetime
import itertools
import logging
import json
import os
import threading
import uuid


# serizalized MetaStorageObject:
# {
#   'update_timestamp': '2023-10-20T15:00:00.000000+00:00' 
#   'filename': '<filename>'
#   'meta': {
#     'key_1': 'abc',
#     'key_2': 2
#   }
# }
class MetaStorageObject:
    VERSION = "1.0"
    def __init__(self, filename, meta=None, update_timestamp=None):
        self._filename = filename
        self._meta = dict() if meta is None else meta
        update_timestamp = datetime.datetime.now(datetime.timezone.utc) if update_timestamp is None else update_timestamp
        self._set_update_timestamp(update_timestamp)

    @staticmethod
    def load_serialized(serialized_data):
        try:
            assert "filename" in serialized_data, "'filename' missing in meta data, data invalid"
            assert "meta" in serialized_data, "'meta' missing in meta data, data invalid"
            assert "update_timestamp" in serialized_data, "'update_timestamp' missing in meta data, data invalid"
            assert serialized_data.get("version", "1.0") == "1.0", f"version {serialized_data.get('version', '1.0')} not supported"
        except AssertionError as e:
            raise MetaStorage.LoadException(f"Error loading meta data '{serialized_data}'", e)
        filename = serialized_data["filename"]
        meta = serialized_data["meta"]
        update_timestamp = datetime.datetime.fromtimestamp((serialized_data["update_timestamp"] / 1000), tz=datetime.timezone.utc)
        return MetaStorageObject(filename, meta, update_timestamp)

    @property
    def filename(self):
        return self._filename

    @filename.setter
    def filename(self, new_filename):
        if new_filename != self._filename:
            self._filename = new_filename
            self._touch()

    @property
    def meta(self):
        return self._meta

    @meta.setter
    def meta(self, new_meta):
        if new_meta == self._meta:
            return
        self._meta = new_meta
        self._touch()

    @property
    def update_timestamp(self):
        return self._update_timestamp

    def _set_update_timestamp(self, new_timestamp):
        assert isinstance(new_timestamp, datetime.datetime)
        assert new_timestamp.tzinfo == datetime.timezone.utc
        self._update_timestamp = new_timestamp.replace(
            microsecond=(new_timestamp.microsecond - (new_timestamp.microsecond % 1000)),
        )

    def _touch(self):
        self._set_update_timestamp(datetime.datetime.now(datetime.timezone.utc))

    def update_meta(self, meta):
        self.meta.update(meta)
        self._touch()

    def delete_meta(self, meta_path):
        MetaStorageObject._recursive_delete_meta_path(self.meta, list(meta_path))
        self._touch()

    @staticmethod
    def _recursive_delete_meta_path(subdict, delete_path):
        if len(delete_path) == 0:
            return True
        delete_key = delete_path.pop(0)
        if delete_key in subdict:
            do_delete = MetaStorageObject._recursive_delete_meta_path(subdict[delete_key], delete_path)
            if do_delete:
                del subdict[delete_key]
                if len(subdict) == 0:
                    return True
        return False

    def serialize(self):
        out = dict(
            filename=self._filename,
            update_timestamp=int(self._update_timestamp.timestamp() * 1000),
            meta=self._meta,
        )
        if self.VERSION != "1.0":
            out.update({"version": self.VERSION})
        return out


class MetaStorageVersionUtil:
    @staticmethod
    def left_version_less_equal(left, right):
        return [int(v) for v in left.split(".")] <= [int(v) for v in right.split(".")]


# serizalized MetaStorage:
# {
#   'data': {
#     '<hash>' : '<MetaStorageObject_serialized>'
#   }
# }
class MetaStorage:
    VERSION="1.0"
    _instances = dict()
    _instances_lock = threading.Lock()

    _meta_extension = ".meta"

    class LoadException(Exception):
        def __init__(self, message, exceptions=[]):
            self.message = message
            if not isinstance(exceptions, (tuple, list)):
                exceptions = [exceptions]
            self.exceptions = exceptions

        def __str__(self):
            linesep = f"{os.linesep}  "
            return f"{self.message}:{linesep}" + linesep.join(itertools.chain(*(str(e).splitlines() for e in self.exceptions)))

    @staticmethod
    def get_abs_directory(path):
        try:
            assert os.path.exists(path), f"{path} does not exist"
        except AssertionError as e:
            raise MetaStorage.LoadException(f"Could not load meta storage for '{path}'", e)

        if os.path.isfile(path):
            path = os.path.dirname(path)
        return os.path.abspath(path)

    def __new__(cls, path):
        abs_directory = MetaStorage.get_abs_directory(path)
        with cls._instances_lock:
            if abs_directory in cls._instances:
                return cls._instances[abs_directory]

            instance = super().__new__(cls)
            instance._initialized = False
            cls._instances[abs_directory] = instance
        return instance

    def __init__(self, path):
        if self._initialized == True: return

        self._initialized = True

        self.path = MetaStorage.get_abs_directory(path)
        self._meta_path = None
        self._data = dict()

        self.load_exception = None
        self._load_meta_data()

    def __len__(self):
        return len(self._data)

    def get_hashes(self):
        return list(self._data.keys())

    def _load_meta_data(self):
        meta_files = self._find_meta_files()

        meta_files_fully_loaded = list()
        meta_file_load_exceptions = list()
        self._data = dict()
        for meta_file in meta_files:
            exception_ocurred = False
            try:
                meta_storage_data = self._load_meta_file(meta_file)
            except Exception as e:
                exception_ocurred = True
                meta_file_load_exceptions.append(
                    MetaStorage.LoadException(f"Could not process file '{meta_file}' for meta data", e)
                )
                continue

            for hash, meta_storage_data_object_data in meta_storage_data["data"].items():
                try:
                    meta_storage_object = MetaStorageObject.load_serialized(meta_storage_data_object_data)
                    if hash in self._data and self._data[hash].update_timestamp > meta_storage_object.update_timestamp:
                        continue
                    self._data[hash] = meta_storage_object
                except Exception as e:
                    meta_file_load_exceptions.append(
                        MetaStorage.LoadException(f"Could not fully process file '{meta_file}' for meta data", e)
                    )
                    exception_ocurred = True
                    continue
            if not exception_ocurred:
                meta_files_fully_loaded.append(meta_file)

        self._meta_path = None
        if len(meta_files) == 1:
            self._meta_path = meta_files[0]
        self._store_data()
        for meta_file in (f for f in meta_files_fully_loaded if f != self._meta_path):
            os.remove(meta_file)


        if len(meta_file_load_exceptions) > 0:
            # raise Exception to safely use logging.Logger.exception
            # According to Python docs (version 3.12.1) for 'logging.Logger.exception':
            # "This method should only be called from an exception handler"
            try:
                raise MetaStorage.LoadException(f"Errors occurred when loading meta storage for path '{self.path}'", meta_file_load_exceptions)
            except MetaStorage.LoadException as e:
                logging.getLogger(__name__).exception(f"Exception occurred when loading meta data")
                self.load_exception = e

    def _find_meta_files(self):
        abs_paths = (os.path.abspath(os.path.join(self.path, file_)) for file_ in os.listdir(self.path) if file_.endswith(self._meta_extension))
        return [path for path in abs_paths if os.path.isfile(path)]

    @staticmethod
    def _load_meta_file(path):
        with open(path, "r") as file_:
            meta_storage_data = json.loads(file_.read())
        try:
            assert "data" in meta_storage_data, "'data' missing in meta storage"
            version = meta_storage_data.get("version", "1.0")
            assert MetaStorageVersionUtil.left_version_less_equal(version, "1.0"), f"metastorage.version={version} not supported"
        except AssertionError as e:
            raise MetaStorage.LoadException(f"Meta file '{path}' corrupt", e)
        return meta_storage_data

    def _store_data(self):
        if len(self._data) == 0:
            if self._meta_path is not None and os.path.exists(self._meta_path):
                os.remove(self._meta_path)
            return

        if self._meta_path is None:
            self._meta_path = os.path.abspath(os.path.join(self.path, f"{uuid.uuid4().hex}{self._meta_extension}"))

        meta_storage_data = dict()
        for hash, meta_storage_object in self._data.items():
            meta_storage_data[hash] = meta_storage_object.serialize()
        
        meta_storage = dict(data=meta_storage_data)
        if self.VERSION != "1.0":
            meta_storage.update({"version": self.VERSION})

        with open(self._meta_path, "w") as file_:
            file_.write(json.dumps(meta_storage))

    def _get_meta_storage_object(self, hash, filename):
        do_store = False
        meta_storage_object = self._data.get(hash, None)
        if meta_storage_object is None:
            meta_storage_object = MetaStorageObject(filename)
            self._data[hash] = meta_storage_object
        if filename is not None and filename != meta_storage_object.filename:
            meta_storage_object.filename = filename
            do_store = True
        if do_store:
            self._store_data()
        return meta_storage_object

    def delete_meta_storage_object(self, hash):
        if hash in self._data:
            del self._data[hash]
        self._store_data()

    def set_meta(self, hash, filename, meta):
        meta_storage_object = self._get_meta_storage_object(hash, filename)
        if meta != meta_storage_object.meta:
            meta_storage_object.meta = meta
            self._store_data()

    def update_meta(self, hash, filename, meta):
        meta_storage_object = self._get_meta_storage_object(hash, filename)
        storage_meta = meta_storage_object.meta
        if any(storage_meta.get(k, object()) != v for k, v in meta.items()):
            meta_storage_object.update_meta(meta)
            self._store_data()

    def delete_meta(self, hash, filename, meta_path=None):
        assert meta_path is None or isinstance(meta_path, (list, tuple))
        meta_storage_object = self._get_meta_storage_object(hash, filename)
        if meta_path is None:
            meta_storage_object.meta = {}
        else:
            meta_storage_object.delete_meta(meta_path)
        self._store_data()

    def get_meta(self, hash, filename=None):
        return self._get_meta_storage_object(hash, filename).meta



# SPDX-FileCopyrightText: 2023 Anthony Zimmermann
#
# SPDX-License-Identifier: GPL-2.0-only

import copy

class RDescriptor:
    @staticmethod
    def not_implemented(name):
        def _not_implemented(owner, *_):
            raise NotImplementedError(f"Function '{name}' of {owner} not implemented!")
        return _not_implemented

    def __set_name__(self, owner, name):
        self._fget_name = f"get_{name}"
        if not hasattr(owner, self._fget_name):
            setattr(owner, self._fget_name, RDescriptor.not_implemented(self._fget_name))

    def __get__(self, obj, objtype=None):
        if obj is None:
            return self
        return getattr(obj, self._fget_name)()


class RWDescriptor(RDescriptor):

    def __set_name__(self, owner, name):
        super().__set_name__(owner, name)
        self._fset_name = f"set_{name}"
        if not hasattr(owner, self._fset_name):
            setattr(owner, self._fset_name, RWDescriptor.not_implemented(self._fset_name))

    def __set__(self, obj, value):
        if obj is None:
            return self
        getattr(obj, self._fset_name)(value)


class SettingsDescriptor:
    def __init__(self, default_value, deserializer=lambda value: value, serializer=lambda value: str(value)):
        self._default_value = default_value
        self._serializer = serializer
        self._deserializer = deserializer
        self._settings_name = None

    def _set_value(self, owner, value):
        if self._settings_name is None:
            raise AttributeError("SettingsDescriptor was not initialized propertly")

        value_serialized = value
        if not isinstance(value, str):
            value_serialized = self._serializer(value_serialized)

        try:
            value_deserialized = self._deserializer(value_serialized)
            assert value_serialized == self._serializer(value_deserialized)
            if not isinstance(value, str):
                assert value == value_deserialized
            else:
                assert value == value_serialized
        except:
            raise ValueError(f"Value '{value}' is incompatible with serializer/deserializer for '{self._settings_name}'")

        settings_dict = getattr(owner, "_settings", dict())
        if getattr(owner, "_settings_owner", None) != owner:
            setattr(owner, "_settings_owner", owner)
            settings_dict = copy.deepcopy(settings_dict)
        settings_dict[self._settings_name] = value_serialized
        setattr(owner, "_settings", settings_dict)

    def __set_name__(self, owner, name):
        self._settings_name = name
        self._set_value(owner, self._default_value)

    def __get__(self, obj, objtype=None):
        if obj is None:
            return self
        if self._settings_name is None:
            raise AttributeError("SettingsDescriptor was not initialized propertly")
        return self._deserializer(getattr(obj, "_settings", dict())[self._settings_name])

    def __set__(self, obj, value):
        self._set_value(obj, value)

class Signal:
    # - observer pattern -
    # This descriptor mimics the behavior of the Qt Signals & Slots.
    # It supports only connecting/disconnecting to/from Python callables.
    # The emit function passes any arguments to the Python callable. The programmer
    # is responsible for matching arguments passed to 'emit' with the signature of all
    # connected callables.
    # Why? The Qt implementation requires inheriting from QObject and calling the
    # constructor for QObject once. This implementation is a solution if multiple
    # inheritance is not desired.
    class SignalConnector:
        def __init__(self, connectable=True):
            self.connectable = connectable
            self._registered_callables = list()

        def connect(self, callable):
            assert self.connectable
            self._registered_callables.append(callable)

        def emit(self, *args, **kwargs):
            for callable in self._registered_callables:
                callable(*args, **kwargs)

    def __set_name__(self, owner, name):
        self._signal_connector_name = name
        # initialize a SignalConnector for the class itself
        if not hasattr(owner, "_signal_connectors"):
            setattr(owner, "_signal_connectors", {self._signal_connector_name: Signal.SignalConnector()})

    def __get__(self, obj, objtype=None):
        if obj is None:
            return self
        # lazy-initialize a SignalConnector on first connect/first emit:
        # <signal_variable>.connect(<callable>) or <signal_variable>.emit(<args>)
        if getattr(obj, "_signal_connectors_owner", None) != obj:
            setattr(obj, "_signal_connectors_owner", obj)
            setattr(obj, "_signal_connectors", {self._signal_connector_name: Signal.SignalConnector()})
        return getattr(obj, "_signal_connectors")[self._signal_connector_name]



# SPDX-FileCopyrightText: 2023 Anthony Zimmermann
#
# SPDX-License-Identifier: GPL-2.0-only

import os
import itertools


class FileSystemScanner:
    @staticmethod
    def find_supported_files(paths, filter_function=lambda _: True):
        def _find_photo_object_files(path):
            photo_object_paths = list()
            if path is None:
                return photo_object_paths
            for root, _, files in os.walk(path):
                root_path = os.path.abspath(root)
                for file in files:
                    abs_path = os.path.join(root_path, file)
                    if filter_function(abs_path) is True:
                        photo_object_paths.append(abs_path)
            return photo_object_paths
        if not isinstance(paths, (list, tuple)):
            paths = (paths,)
        return list(sorted(itertools.chain.from_iterable(_find_photo_object_files(p) for p in paths)))



# SPDX-FileCopyrightText: 2023 Anthony Zimmermann
#
# SPDX-License-Identifier: GPL-2.0-only

import contextlib

from PySide6.QtCore import QAbstractListModel
from PySide6.QtCore import Qt
from PySide6.QtCore import QModelIndex


class PhotoDataModel(QAbstractListModel):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.photo_index = []

    def set_photo_index(self, photo_index):
        self.beginResetModel()
        self.photo_index = photo_index
        self.endResetModel()

    @contextlib.contextmanager
    def updating_photo(self, photo):
        if photo not in self.photo_index:
            yield
            return
        index = self.photo_index.index(photo)
        try:
            yield
        finally:
            qmodel_index = self.index(index, index, QModelIndex())
            self.dataChanged.emit(qmodel_index, qmodel_index, [Qt.ItemDataRole.UserRole])

    @contextlib.contextmanager
    def removing_photo(self, photo):
        if photo not in self.photo_index:
            yield
            return
        index = self.photo_index.index(photo)
        try:
            self.beginRemoveRows(QModelIndex(), index, index)
            yield
        finally:
            self.endRemoveRows()

    def rowCount(self, parent=None):
        return len(self.photo_index)

    def data(self, index, role=Qt.ItemDataRole.DisplayRole):
        photo_object = self.photo_index[index.row()]
        # (Extracted from source code documentation at
        # src/widgets/intemviews/qstyleditemdelegate.cpp):
        # Role                             | Accepted Types
        # ---------------------------------+------------------------------------
        # # Qt::AccessibleDescriptionRole  | QString
        # # Qt::AccessibleTextRole         | QString
        # Qt::BackgroundRole               | QBrush
        # Qt::CheckStateRole               | Qt::CheckState
        # Qt::DecorationRole               | QIcon, QPixmap, QImage and QColor
        # Qt::DisplayRole                  | QString and types with a string
        #                                    representation
        # Qt::EditRole                     | See QItemEditorFactory for details
        # Qt::FontRole                     | QFont
        # Qt::SizeHintRole                 | QSize
        # # Qt::StatusTipRole              | ? 
        # Qt::TextAlignmentRole            | Qt::Alignment
        # Qt::ForegroundRole               | QBrush
        # # Qt::ToolTipRole                | ?
        # # Qt::WhatsThisRole              | ?

        # This model uses Qt::UserRole to pass a PhotoObject
        if role == Qt.ItemDataRole.DisplayRole:
            return photo_object.filename
        if role == Qt.ItemDataRole.UserRole:
            return photo_object
        # TODO other roles
        return None



# SPDX-FileCopyrightText: 2023 Anthony Zimmermann
#
# SPDX-License-Identifier: GPL-2.0-only

import importlib.resources
import logging
import math
import threading
import time

from PySide6.QtCore import QSize
from PySide6.QtCore import Qt
from PySide6.QtCore import QTimer
from PySide6.QtGui import QFontMetrics
from PySide6.QtGui import QImageReader
from PySide6.QtGui import QPixmap
from PySide6.QtGui import QTransform
from PySide6.QtGui import QKeySequence
from PySide6.QtWidgets import QComboBox
from PySide6.QtWidgets import QDialog
from PySide6.QtWidgets import QFrame
from PySide6.QtWidgets import QGridLayout
from PySide6.QtWidgets import QGroupBox
from PySide6.QtWidgets import QLabel
from PySide6.QtWidgets import QListWidget
from PySide6.QtWidgets import QMessageBox
from PySide6.QtWidgets import QPushButton
from PySide6.QtWidgets import QProgressBar
from PySide6.QtWidgets import QSizePolicy
from PySide6.QtWidgets import QTextEdit
from PySide6.QtWidgets import QWidget

from photoassistant.utils.descriptors import RWDescriptor
from photoassistant.utils.descriptors import Signal


class SynchronizedExecution:
    @staticmethod
    def execute_event_blocked(function, event=None, polling_interval_ms=100):
        timer = QTimer()
        if event is None:
            event = threading.Event()
        def _try_execute():
            event.wait(timeout=0.001)
            if event.is_set():
                function()
                timer.stop()
        timer.timeout.connect(_try_execute)
        timer.start(polling_interval_ms)
        return event

    @staticmethod
    def execute_delayed(function, delay_ms=0):
        QTimer.singleShot(delay_ms, function)


class QKeySequenceStandardKeyUtils:
    @staticmethod
    def get_user_string(q_standard_key):
        user_string = QKeySequence(q_standard_key).toString().upper()
        def replace_keys(user_string_part, replacements):
            for orig, repl in replacements.items():
                if user_string_part.startswith(orig):
                    user_string_part = repl + user_string_part[len(orig):]
                    break
            parts = user_string_part.split("+", 1)
            if len(parts) == 1:
                return parts[0]
            return "+".join((parts[0], replace_keys(parts[1], replacements)))
        return replace_keys(
            user_string,
            {
                "+": "PLUS",
                "-": "MINUS"
            },
        )


class QSizePolicyStretch(QSizePolicy):
    def __init__(self, horizontal, vertical, stretch_horizontal=1, stretch_vertical=1):
        super().__init__(horizontal, vertical)
        self.setHorizontalStretch(stretch_horizontal)
        self.setVerticalStretch(stretch_vertical)


class QComboBoxEditable(QComboBox):
    def __init__(self, minimum_size=(100, 20), **kwargs):
        super().__init__(**kwargs)
        self.setEditable(True)
        self.setMinimumSize(*minimum_size)


class QListWidgetTransparentMinimumSized(QListWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.viewport().setAutoFillBackground(False)
        self.setFrameShape(QFrame.Shape.NoFrame)

    def minimumSizeHint(self):
        return self.minimumSize()


class QTextEditResizing(QTextEdit):
    def __init__(self, parent=None, max_height_lines=4.1):
        super().__init__(parent)
        self._max_height_lines = max_height_lines

        self.setContentsMargins(0, 0, 0, 0)
        self.document().setDocumentMargin(0.0)

        size_policy = self.sizePolicy()
        size_policy.setHeightForWidth(True)
        size_policy.setVerticalPolicy(QSizePolicy.Policy.Fixed)
        self.setSizePolicy(size_policy)

        self.textChanged.connect(self.updateGeometry)

    def resizeEvent(self, event):
        contents_margins = self.contentsMargins()
        self.setMinimumWidth(50 + contents_margins.left() + contents_margins.right())
        self.updateGeometry()
        super().resizeEvent(event)

    def hasHeightForWidth(self):
        return True

    def heightForWidth(self, width):
        margins = self.contentsMargins()

        document_width = max(0, (width - margins.left() - margins.right()))
        document = self.document().clone()
        document.setTextWidth(document_width)
        full_height = math.ceil(margins.top() + document.size().height() + margins.bottom())
        max_height = math.ceil(
            margins.top()
            + (2 * document.documentMargin())
            + margins.bottom()
            + (self._max_height_lines * QFontMetrics(document.defaultFont()).lineSpacing())
        )
        return min(full_height, max_height)

    def minimumSizeHint(self):
        # self.ensurePolished()  # this function is implicitly called in super().sizeHint()
        width = super().sizeHint().width()
        return QSize(width, self.heightForWidth(self.width()))

    def sizeHint(self):
        return self.minimumSizeHint()


class QModernStyleGroupBox(QGroupBox):
    def __init__(self, title, parent=None):
        super().__init__(title, parent)
        self.setStyleSheet(
            "QGroupBox {"
            "border: 1px solid gray;"
            "border-radius: 0.5ex;"
            "margin-top: 0.6em;"
            "padding: 0.4em 2px 2px 2px;"
            "background-color: transparent;"
            "}"
            "QGroupBox::title {"
            "subcontrol-origin: margin;"
            "subcontrol-position: top left;"
            "padding: 0 3px 0 3px;"
            "left: 1.5ex;"
            "}"
        )


class QModernStyleGroupBoxContainer(QModernStyleGroupBox):
    def __init__(self, title, widget=None, parent=None):
        super().__init__(title, parent)

        self._grid_layout = QGridLayout(self)
        self._grid_layout.setContentsMargins(0, 0, 0, 0)

        if widget is not None:
            self._grid_layout.addWidget(widget)


class QLabeledProperty(QModernStyleGroupBoxContainer):
    class QLabeledPropertyLabel(QTextEditResizing):
        def __init__(self, parent=None, max_height_lines=4.1):
            super().__init__(parent, max_height_lines)

            self.setReadOnly(True)
            self.setFrameStyle(QFrame.Shape.NoFrame)
            self.viewport().setAutoFillBackground(False)
            self.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
            self.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)

        def wheelEvent(self, event):
            if not (self.textInteractionFlags() & Qt.TextInteractionFlag.TextEditable):
                if event.modifiers() == Qt.KeyboardModifier.ControlModifier:
                    return
            super().wheelEvent(event)

        def focusOutEvent(self, event):
            cursor = self.textCursor()
            cursor.clearSelection()
            self.setTextCursor(cursor)
            super().focusOutEvent(event)


    def __init__(self, title, parent=None):
        super().__init__(title, parent)

        self.setFlat(True)
        self._property_label = self.QLabeledPropertyLabel()

        self.layout().addWidget(self._property_label, 0, 0)

    def set_text(self, text):
        self._property_label.setText(text)


class QMarginContainer(QWidget):
    def __init__(self, parent=None, widget=None, margin=(0, 0, 0, 0)):
        super().__init__(parent)

        self.setLayout(QGridLayout())

        self._widget = None
        self.set_margin(*margin)
        self.set_widget(widget)

    def set_margin(self, left, top, right, bottom):
        self.layout().setContentsMargins(left, top, right, bottom)

    def set_widget(self, widget):
        if widget == self._widget:
            return
        if self._widget is not None:
            self._widget.setParent(None)
        self._widget = widget
        self.layout().addWidget(self._widget, 0, 0)


class QPixmapLoader:
    @staticmethod
    def load(path, max_size=None, orientation_correction=None):
        try:
            reader = QImageReader(path)
            if max_size is not None:
                reader.setScaledSize(
                    reader.size().scaled(
                        max_size,
                        Qt.AspectRatioMode.KeepAspectRatio,
                    )
                )
            image = reader.read()
        except Exception:
            logging.getLogger(__name__).exception(f"Processing file '{path}' for image data failed")
            with importlib.resources.path("photoassistant.gui.resources", "icon-broken-image.png") as icon_path:
                return QPixmapLoader.load(str(icon_path.resolve()))

        if orientation_correction is not None:
            # We don't need to implement transformation into center and back
            # around the orientation_correction as the function QImage.transformed
            # will implicitly correct the transformation matrix to compensate for
            # transformations that move the center of the resulting image.
            q_transform = QTransform(
                orientation_correction[0],
                orientation_correction[3],
                orientation_correction[6],
                orientation_correction[1],
                orientation_correction[4],
                orientation_correction[6],
                orientation_correction[2],
                orientation_correction[5],
                orientation_correction[8],
            )
            image = image.transformed(q_transform)
        return QPixmap.fromImage(image)


class SimpleQMessageBox(QMessageBox):
    ICON = None
    BUTTON = QMessageBox.StandardButton.Ok
    def __init__(self, parent, message, title="INFO"):
        super().__init__(parent)
        self.setWindowTitle(title)
        self.setText(message)
        self.setStandardButtons(self.BUTTON)
        if self.ICON is not None:
            self.setIcon(self.ICON)


class InfoQMessageBox(SimpleQMessageBox):
    ICON = QMessageBox.Icon.Information


class ErrorQMessageBox(SimpleQMessageBox):
    def __init__(self, parent, message, title="ERROR"):
        super().__init__(parent, message, title)
    ICON = QMessageBox.Icon.Critical


class ScrollableInfoDialog(QDialog):
    def __init__(self, parent=None, message="", title="PhotoAssistant"):
        super().__init__(parent)

        self.text_edit = QTextEdit()
        self.text_edit.setMarkdown(message)
        self.text_edit.setReadOnly(True)

        close_button = QPushButton("close")

        close_button_layout = QGridLayout()
        close_button_layout.addWidget(close_button, 0, 1)
        close_button_layout.setColumnStretch(0, 1)
        close_button_layout.setColumnStretch(1, 1)
        close_button_layout.setColumnStretch(2, 1)

        close_button.clicked.connect(self.hide)

        grid_layout = QGridLayout(self)
        grid_layout.addWidget(self.text_edit, 0, 0)
        grid_layout.addLayout(close_button_layout, 1, 0)

    def set_text(self, text, title="PhotoAssistant"):
        self.setWindowTitle(title)
        self.text_edit.setText(text)
        self.show()

    def set_markdown(self, text, title="PhotoAssistant"):
        self.setWindowTitle(title)
        self.text_edit.setMarkdown(text)
        self.show()


class BusyIndicatorDialog(QDialog):
    SHOW_DELAY = 0.5 # in seconds

    class ExecuteResult:
        NONE = object()
        data = RWDescriptor()

        def __init__(self, data=NONE):
            self._data_lock = threading.Lock()
            self._data_ready_event = threading.Event()
            self._data = None
            if data is not BusyIndicatorDialog.ExecuteResult.NONE:
                self.set_data(data)

        def get_data(self):
            with self._data_lock:
                return self._data

        def set_data(self, data):
            with self._data_lock:
                self._data = data
            self._data_ready_event.set()

        def is_ready(self):
            return self._data_ready_event.is_set()

        def wait(self, *args, **kwargs):
            self._data_ready_event.wait(*args, **kwargs)
            return self.data

    class MessageUpdateModule(threading.Thread):
        update_signal = Signal()
        MINIMUM_MESSAGE_DISPLAY_TIME = 3 # in seconds

        def __init__(self, start_delay, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.start_delay = start_delay

            self._message_queue = []
            self._message_queue_lock = threading.Lock()
            self._message_event = threading.Event()
            self._stop_event = threading.Event()

        def run(self):
            time.sleep(self.start_delay)
            while True:
                self._message_event.wait()
                if self._stop_event.is_set():
                    self._cleanup()
                    return
                with self._message_queue_lock:
                    if len(self._message_queue) == 0:
                        self._message_event.clear()
                        continue
                    self.update_signal.emit(self._message_queue.pop(0))
                time.sleep(self.MINIMUM_MESSAGE_DISPLAY_TIME)

        def stop(self):
            # First, set _stop_event such that the thread loop (run) will call
            # _cleanup() when _message_event is set.
            # Afterwards, set _message_event to release the
            # _message_event.wait which is maybe currently being executed.
            self._stop_event.set()
            self._message_event.set() # release wait on _message_event

        def add_message(self, message):
            with self._message_queue_lock:
                self._message_queue.append(message)
            self._message_event.set()

        def _cleanup(self):
            # delete all messages from the _message_queue and clear events
            with self._message_queue_lock:
                self._message_queue = []
            self._stop_event.clear()
            self._message_event.clear()

    def __init__(self, parent):
        super().__init__(parent)
        self.setWindowTitle("PhotoAssistant is running an operation that might take a while ...")
        self.setModal(True)

        self.info_label = QLabel()

        self.progress_bar = QProgressBar()
        self.progress_bar.setMinimum(0)
        self.progress_bar.setMaximum(0)

        grid_layout = QGridLayout(self)
        grid_layout.addWidget(self.info_label)
        grid_layout.addWidget(self.progress_bar)

        self.execute_thread_lock = threading.RLock()
        self.execute_thread = None
        self.message_update_module = None

        self.block_closing = True

    def execute(self, message, func, *args, **kwargs):
        # start one thread which picks up the messages to show and displays them
        # for at least MINIMUM_MESSAGE_DISPLAY_TIME in seconds
        with self.execute_thread_lock:
            if self.execute_thread is not None:
                self.set_message(message)
                return self.ExecuteResult(func(*args, **kwargs))

        self.info_label.setText(message)
        execute_result = self.ExecuteResult()
        def _target():
            with self.execute_thread_lock:
                self.message_update_module = self.MessageUpdateModule(self.SHOW_DELAY)
                self.message_update_module.update_signal.connect(self._show_message)

                self.message_update_module.start()
                execute_result.data = func(*args, **kwargs)
                self.message_update_module.stop()

                # clear references to execute_thead and message_update_module
                # such that the BusyIndicatorDialog is ready for starting the next
                # execution
                self.execute_thread = None
                self.message_update_module = None
        self.execute_thread = threading.Thread(target=_target)
        self.execute_thread.start()
        self._show_dialog(execute_result._data_ready_event)
        return execute_result

    def _show_message(self, message):
        self.info_label.setText(message)

    def _show_dialog(self, execute_result_ready_event):
        def _show_dialog_if_still_executing():
            if self.execute_thread is not None:
                super(self.__class__, self).show()
                SynchronizedExecution.execute_event_blocked(super(self.__class__, self).hide, event=execute_result_ready_event)
        SynchronizedExecution.execute_delayed(_show_dialog_if_still_executing, delay_ms=int(self.SHOW_DELAY * 1000))

    def set_message(self, message):
        if self.message_update_module is not None:
            self.message_update_module.add_message(message)

    # from PySide6 documentation: "In order to modify your dialog’s close behavior,
    # you can reimplement the functions accept() , reject() or done()"
    def accept(self):
        if not self.block_closing: super().accept()

    # from PySide6 documentation: "In order to modify your dialog’s close behavior,
    # you can reimplement the functions accept() , reject() or done()"
    def reject(self):
        if not self.block_closing: super().reject()

    # from PySide6 documentation: "In order to modify your dialog’s close behavior,
    # you can reimplement the functions accept() , reject() or done()"
    def done(self, result):
        if not self.block_closing: super().done(result)



# SPDX-FileCopyrightText: 2023 Anthony Zimmermann
#
# SPDX-License-Identifier: GPL-2.0-only

import datetime
import exifread
import hashlib
import logging
import mmap
import os
import re

from photoassistant.core.metastorage import MetaStorage
from photoassistant.utils.cachingutils import AsyncSharedCache
from photoassistant.utils.descriptors import RDescriptor
from photoassistant.utils.descriptors import RWDescriptor
from photoassistant.utils.decoratorutils import async_cached_query

# exifread prints a warning 'PNG file does not have exif data.' when exif data
# is not present. This workaround is to get rid of the warning.
logging.getLogger("exifread").setLevel(logging.ERROR)

class PhotoHash:
    @staticmethod
    def sha256(path):
        hash_object = hashlib.sha256()
        with open(path, "rb") as file_:
            with mmap.mmap(file_.fileno(), 0, access=mmap.ACCESS_READ) as map:
                hash_object.update(map)
        return hash_object.hexdigest()


class PhotoInterface:
    path = RWDescriptor()
    filename = RDescriptor()
    main_tag = RWDescriptor()
    tags = RDescriptor()
    all_tags = RDescriptor()
    creation_timestamp = RWDescriptor()
    orientation_correction = RWDescriptor()
    gps_coordinates = RWDescriptor()
    hash = RDescriptor()
    image = RDescriptor()

    def __init__(self, path):
        path = os.path.abspath(path)
        assert self.file_is_supported(path)
        self._path = path

    def get_path(self):
        return self._path

    def get_filename(self):
        return os.path.basename(self._path)

    def add_tag(self, tag):
        raise NotImplementedError()

    def delete_tag(self, tag):
        raise NotImplementedError()

    @staticmethod
    def file_is_supported(file):
        raise NotImplementedError()


class Photo(PhotoInterface):
    meta = RDescriptor()
    exif = RDescriptor()

    class TagFormatException(Exception):
        pass

    TAG_ALLOWED_CHARACTERS = r"[a-zA-Z0-9_+\-*/#'!%~\(\)&?\.,;]"
    MAIN_TAG_REGEX = re.compile(r"(?:\:" + TAG_ALLOWED_CHARACTERS + "+)+")
    TAG_REGEX = re.compile("^" + TAG_ALLOWED_CHARACTERS + "+$")
    # mapping of exif id to correction matrices for images centered in the origin
    EXIF_ORIENTATION_CORRECTION_MATRIX_MAP = {
        1: [1, 0, 0, 0, 1, 0, 0, 0, 1],
        2: [-1, 0, 0, 0, 1, 0, 0, 0, 1],
        3: [-1, 0, 0, 0, -1, 0, 0, 0, 1],
        4: [1, 0, 0, 0, -1, 0, 0, 0, 1],
        5: [0, 1, 0, 1, 0, 0, 0, 0, 1],
        6: [0, 1, 0, -1, 0, 0, 0, 0, 1],
        7: [0, -1, 0, -1, 0, 0, 0, 0, 1],
        8: [0, -1, 0, 1, 0, 0, 0, 0, 1],
    }

    def get_exif(self):
        with open(self.get_path(), "rb") as file_:
            with mmap.mmap(file_.fileno(), 0, access=mmap.ACCESS_READ) as mmap_:
                try:
                    return exifread.process_file(mmap_, details=False)
                except Exception:
                    logging.getLogger(__name__).exception(f"Processing file '{self.get_path()}' for exif data failed")
                    return dict()

    def get_meta(self):
        return MetaStorage(self.get_path()).get_meta(self.get_hash(), self.get_filename())

    def get_all_tags(self):
        return list(self.get_meta().get("tags", []))

    def get_tags(self):
        return [t for t in self.get_all_tags() if not t.startswith(":")]

    def get_main_tag(self):
        for tag in filter(lambda t: t.startswith(":"), self.get_all_tags()):
            return tag  # return first tag if present
        return None

    def set_main_tag(self, tag):
        if self.get_main_tag() != tag:
            all_tags = self.get_tags()
            if tag is not None:
                try:
                    assert self.MAIN_TAG_REGEX.match(tag)
                except AssertionError:
                    raise Photo.TagFormatException(f"Tag '{tag}' is ill-formatted")
                all_tags = [tag] + all_tags
            MetaStorage(self.get_path()).update_meta(self.get_hash(), self.get_filename(), meta={"tags": all_tags})

    def add_tag(self, tag):
        try:
            assert self.TAG_REGEX.match(tag)
        except AssertionError:
            raise Photo.TagFormatException(f"Tag '{tag}' is ill-formatted")
        if tag not in self.get_tags():
            all_tags = self.get_all_tags() + [tag]
            MetaStorage(self.get_path()).update_meta(self.get_hash(), self.get_filename(), meta={"tags": all_tags})

    def delete_tag(self, tag):
        if tag in self.get_tags():
            all_tags = self.get_all_tags()
            all_tags.remove(tag)
            MetaStorage(self.get_path()).update_meta(self.get_hash(), self.get_filename(), meta={"tags": all_tags})

    def get_creation_timestamp(self):
        meta_timestamp = self.get_meta().get("creation_timestamp")
        if meta_timestamp is not None:
            return datetime.datetime.fromtimestamp(meta_timestamp / 1000)

        exif_timestamp = self.get_exif().get("Image DateTime")
        if exif_timestamp is not None:
            return datetime.datetime.strptime(exif_timestamp.values, "%Y:%m:%d %H:%M:%S")

        return None

    def set_creation_timestamp(self, dt):
        assert isinstance(dt, datetime.datetime)
        assert dt.tzinfo is None
        MetaStorage(self.get_path()).update_meta(self.get_hash(), self.get_filename(), meta={"creation_timestamp": int(dt.timestamp() * 1000)})

    def _get_exif_orientation_correction(self):
        exif_orientation = self.get_exif().get("Image Orientation")
        if exif_orientation is not None:
            exif_orientation = exif_orientation.values[0]
            # Even though only values 1 to 8 are valid exif orientation meta data values,
            # some cameras use e.g., 0 for panorama images. Hence, we gracefully assume
            # 'no orientation' when the value is not between [1, 8].
            if 1 <= exif_orientation <= 8:
                return Photo.EXIF_ORIENTATION_CORRECTION_MATRIX_MAP[exif_orientation]
        return [1, 0, 0, 0, 1, 0, 0, 0, 1]

    def get_orientation_correction(self):
        meta_orientation = self.get_meta().get("orientation_correction")
        if meta_orientation is not None:
            return meta_orientation

        return self._get_exif_orientation_correction()

    def set_orientation_correction(self, orientation_correction):
        if (
            not isinstance(orientation_correction, (tuple, list))
            or len(orientation_correction) != 9
        ):
            raise ValueError("orientation_correction must be a list of 9 entries representing the values of a a 3x3 matrix")
        exif_orientation_correction = self._get_exif_orientation_correction()
        if exif_orientation_correction is not None and all(a == b for a, b in zip(orientation_correction, exif_orientation_correction)):
            MetaStorage(self.get_path()).delete_meta(self.get_hash(), self.get_filename(), ["orientation_correction"])
        else:
            MetaStorage(self.get_path()).update_meta(self.get_hash(), self.get_filename(), meta={"orientation_correction": orientation_correction})

    def _get_exif_gps_coordinates(self):
        def _get_exif_degrees_generic(orientation):
            coordinate = self.get_exif().get(f"GPS GPS{orientation}")
            coordinate_ref = self.get_exif().get(f"GPS GPS{orientation}Ref")
            if coordinate is None or coordinate_ref is None:
                return None
            d, m, s = coordinate.values
            try:
                coordinate_degrees = d + (m / 60) + (s / 3600)
                if coordinate_ref.values in ("S", "W"):
                    coordinate_degrees *= -1
            except ZeroDivisionError:
                return None
            return float(coordinate_degrees)

        def _get_altitude():
            value = self.get_exif().get(f"GPS GPSAltitude")
            value_ref = self.get_exif().get(f"GPS GPSAltitudeRef")
            if value is None or value_ref is None:
                return None
            value = value.values[0]
            if value_ref.values == 1:
                value *= -1
            return float(value)

        latitude = _get_exif_degrees_generic("Latitude")
        longitude = _get_exif_degrees_generic("Longitude")
        altitude = _get_altitude()
        meta_gps_coordinates = [latitude, longitude, altitude]
        return meta_gps_coordinates

    def get_gps_coordinates(self):
        meta_gps_coordinates = self.get_meta().get("gps_coordinates")
        if meta_gps_coordinates is not None:
            return meta_gps_coordinates
        return self._get_exif_gps_coordinates()

    def set_gps_coordinates(self, gps_coordinates):
        if (
            not isinstance(gps_coordinates, (tuple, list))
            or len(gps_coordinates) != 3
        ):
            raise ValueError("gps_coordinates must be a list of 3 entries representing latitude, longitude and altitude")
        exif_gps_coordinates = self._get_exif_gps_coordinates()
        if exif_gps_coordinates is not None and all(a == b for a, b in zip(gps_coordinates, exif_gps_coordinates)):
            MetaStorage(self.get_path()).delete_meta(self.get_hash(), self.get_filename(), ["gps_coordinates"])
        else:
            MetaStorage(self.get_path()).update_meta(self.get_hash(), self.get_filename(), meta={"gps_coordinates": gps_coordinates})

    def get_hash(self):
        return PhotoHash.sha256(self.get_path())


class PhotoAsyncSharedCached(Photo):
    ASYNC_SHARED_CACHE_EXIF = AsyncSharedCache(slots=100000)
    ASYNC_SHARED_CACHE_HASH = AsyncSharedCache(slots=100000)

    #override
    @async_cached_query(ASYNC_SHARED_CACHE_EXIF)
    def get_exif(self, **kwargs):
        # Special keyword arguments:
        # 'query_ignore_cache':       default=False - set to False if cache should be fully ignored
        # 'query_cache_state_only':   default=False - do not execute the query, only peek into cache if value is available
        # 'query_callback':           default=None - sets a callback to be called when the result is available
        # 'query_result_placeholder': default=AsyncSharedCacheQueryPolicy.NONE - value to be returned if the value is not yet cached
        return super().get_exif()

    #override
    @async_cached_query(ASYNC_SHARED_CACHE_HASH)
    def get_hash(self, **kwargs):
        # Special keyword arguments:
        # 'query_ignore_cache':       default=False - set to False if cache should be fully ignored
        # 'query_cache_state_only':   default=False - do not execute the query, only peek into cache if value is available
        # 'query_callback':           default=None - sets a callback to be called when the result is available
        # 'query_result_placeholder': default=AsyncSharedCacheQueryPolicy.NONE - value to be returned if the value is not yet cached
        return super().get_hash()



# SPDX-FileCopyrightText: 2023 Anthony Zimmermann
#
# SPDX-License-Identifier: GPL-2.0-only

import math

from PySide6.QtCore import QItemSelection
from PySide6.QtCore import QModelIndex
from PySide6.QtCore import QPoint
from PySide6.QtCore import QRect
from PySide6.QtCore import QRectF
from PySide6.QtCore import QSize
from PySide6.QtCore import Qt
from PySide6.QtCore import Signal as QSignal
from PySide6.QtGui import QKeySequence
from PySide6.QtGui import QPainter
from PySide6.QtGui import QRegion
from PySide6.QtWidgets import QAbstractItemView
from PySide6.QtWidgets import QStyle
from PySide6.QtWidgets import QStyleOptionViewItem
from PySide6.QtWidgets import QStylePainter

from photoassistant.utils.cachingutils import CacheState
from photoassistant.utils.delayutils import DelayedQueryManager
from photoassistant.gui.qt.photoviewitemdelegate import PhotoViewItemDelegate
from photoassistant.gui.qt.utils.qtutils import QKeySequenceStandardKeyUtils

class PhotoView(QAbstractItemView):
    def __init__(self, parent=None):
        super().__init__(parent)

        self._photo_tiles_per_row = 4
        self.photo_tiles_visible = (self.photo_tiles_per_row * 1)
        self.photo_tile_width = 100
        self.photo_tile_height = 100
        self.rect_photo_grid = None
        self.update_photo_tile_size()

        self.setFocusPolicy(Qt.FocusPolicy.WheelFocus)
        self.horizontalScrollBar().setRange(0, 0)
        self.verticalScrollBar().setRange(0, 0)
        self.setItemDelegate(PhotoViewItemDelegate(self))

        self.pixmap_query_manager = DelayedQueryManager()
        self.pixmap_query_manager.set_max_queries(100)

    @property
    def photo_tiles_per_row(self):
        return self._photo_tiles_per_row

    @photo_tiles_per_row.setter
    def photo_tiles_per_row(self, photo_tiles_per_row):
        if 0 < photo_tiles_per_row <= 10:
            self._photo_tiles_per_row = photo_tiles_per_row
            self.reset()

    def setItemDelegate(self, *args, **kwargs):
        super().setItemDelegate(*args, **kwargs)
        self.update_photo_tile_size()

    def update_photo_tile_size(self):
        self.photo_tile_width = (self.viewport().width() / self.photo_tiles_per_row)
        self.photo_tile_height = self.photo_tile_width
        if hasattr(self.itemDelegate(), "photo_view_text_height"):
            self.photo_tile_height += getattr(self.itemDelegate(), "photo_view_text_height")
        photo_tile_width = max(1, self.photo_tile_width)
        photo_tile_height = max(1, self.photo_tile_height)
        self.photo_load_max_size = QSize((1 << math.ceil(math.log2(photo_tile_width))), (1 << math.ceil(math.log2(photo_tile_height))))
        photo_tiles_per_column = math.ceil(self.viewport().height() / photo_tile_height)
        self.photo_tiles_visible = (self.photo_tiles_per_row * photo_tiles_per_column)

        model = self.model()
        if model is not None:
            model_index_0 = model.index(0, 0, self.rootIndex())
            photo_0 = model_index_0.data(Qt.ItemDataRole.UserRole)
            if hasattr(photo_0, "ASYNC_SHARED_CACHE_IMAGES"):
                photo_0.ASYNC_SHARED_CACHE_IMAGES.set_slots(10 * self.photo_tiles_visible)

    def update_rect_photo_grid(self):
        rect_last_photo_tile = self.rect_for_row((self.model().rowCount() - 1)).toRect()
        self.rect_photo_grid = QRect(0, 0, self.viewport().width(), rect_last_photo_tile.bottom())

    def visualRect(self, index):
        return self.viewport_rect_for_row(index.row()).toRect()

    def rect_for_row(self, row):
        photo_grid_row_index = int(row / self.photo_tiles_per_row)
        photo_grid_col_index = (row % self.photo_tiles_per_row)
        return QRectF(
            (photo_grid_col_index * self.photo_tile_width),
            (photo_grid_row_index * self.photo_tile_height),
            self.photo_tile_width,
            self.photo_tile_height,
        )

    def viewport_rect_for_row(self, row):
        rect = self.rect_for_row(row)
        return rect.translated(
            -self.horizontalScrollBar().value(),
            -self.verticalScrollBar().value(),
        )

    def isIndexHidden(self, index):
        return False # TODO ?

    def scrollTo(self, index, scroll_hint):
        view_rect = self.viewport().rect()
        item_rect = self.visualRect(index)

        if item_rect.left() < view_rect.left():
            offset = item_rect.left() - view_rect.left()
            self.horizontalScrollBar().setValue(self.horizontalScrollBar().value() + offset)
        elif item_rect.right() > view_rect.right():
            offset = min(
                item_rect.right() - view_rect.right(),
                item_rect.left() - view_rect.left(),
            )
            self.horizontalScrollBar().setValue(self.horizontalScrollBar().value() + offset)

        if item_rect.top() < view_rect.top():
            offset = item_rect.top() - view_rect.top()
            self.verticalScrollBar().setValue(self.verticalScrollBar().value() + offset)
        elif item_rect.bottom() > view_rect.bottom():
            offset = min(
                item_rect.bottom() - view_rect.bottom(),
                item_rect.top() - view_rect.top(),
            )
            self.verticalScrollBar().setValue(self.verticalScrollBar().value() + offset)

        self.viewport().update()

    def photo_grid_position_at(self, point):
        photo_grid_row_index = int(point.y() / self.photo_tile_height)
        photo_grid_col_index = int(point.x() / self.photo_tile_width)
        return (photo_grid_col_index, photo_grid_row_index)

    def row_of_photo_grid_position(self, photo_grid_col_index, photo_grid_row_index):
        return (photo_grid_row_index * self.photo_tiles_per_row) + photo_grid_col_index

    def indexAt(self, point):
        photo_grid_point = QPoint(
            (point.x() + self.horizontalScrollBar().value()),
            (point.y() + self.verticalScrollBar().value()),
        )
        photo_grid_col_index, photo_grid_row_index = self.photo_grid_position_at(photo_grid_point)
        row = self.row_of_photo_grid_position(photo_grid_col_index, photo_grid_row_index)
        return self.model().index(row, 0, self.rootIndex())

    def moveCursor(self, cursor_action, modifiers):
        current_index = self.currentIndex()
        if not current_index.isValid():
            return QModelIndex()

        current_row = current_index.row()
        max_row = (self.model().rowCount() - 1)
        new_row = None

        if cursor_action == QAbstractItemView.CursorAction.MoveLeft and current_row > 0:
            new_row = (current_row - 1)
        if cursor_action == QAbstractItemView.CursorAction.MoveUp and current_row > 0:
            new_row = (current_row - self.photo_tiles_per_row)

        if cursor_action == QAbstractItemView.CursorAction.MoveRight and current_row < max_row:
            new_row = (current_row + 1)
        if cursor_action == QAbstractItemView.CursorAction.MoveDown and current_row < max_row:
            new_row = (current_row + self.photo_tiles_per_row)

        if new_row is None:
            return QModelIndex()

        new_row = max(0, min(max_row, new_row))
        return self.model().index(new_row, 0, self.rootIndex())

    def horizontalOffset(self):
        return self.horizontalScrollBar().value()

    def verticalOffset(self):
        return self.verticalScrollBar().value()
    
    def scrollContentsBy(self, dx, dy):
        self.scrollDirtyRegion(dx, dy)
        self.viewport().scroll(dx, dy)

    def setSelection(self, rect, flags):
        item_selection = QItemSelection()
        item_selection.select(
            self.indexAt(rect.topLeft()),
            self.indexAt(rect.bottomRight()),
        )
        self.selectionModel().select(item_selection, flags)

    def visualRegionForSelection(self, selection):
        region = QRegion()
        for item_selection_range in selection:
            for row in range(item_selection_range.top(), item_selection_range.bottom()):
                for column in range(item_selection_range.left(), item_selection_range.right()):
                    region += self.visualRect(
                        self.model().index(row, column, self.rootIndex()),
                    )
        return region

    def _preload_photo_image(self, row):
        model = self.model()
        index = model.index(row, 0, self.rootIndex())
        if not index.isValid():
            return
        photo = index.data(Qt.ItemDataRole.UserRole)
        cache_state = photo.get_image(max_size=self.photo_load_max_size, query_cache_state_only=True)
        if cache_state == CacheState.DATA_MISSING:
            # We need to issue a query once with a callback as soon as the image is ready

            def query_callback(_):
                model.dataChanged.emit(index, index, [Qt.ItemDataRole.UserRole])
                self.pixmap_query_manager.stop_query_if_pending(photo)

            self.pixmap_query_manager.start_query_if_none_pending(
                photo,
                lambda: photo.get_image(
                    max_size=self.photo_load_max_size,
                    query_callback=query_callback,
                    query_result_placeholder=None,
                )
            )

    def paintEvent(self, event):
        painter = QStylePainter(self.viewport())
        painter.setRenderHints(QPainter.RenderHint.Antialiasing | QPainter.RenderHint.TextAntialiasing)
        photo_grid_rect = self.viewport().rect().translated(
            self.horizontalScrollBar().value(),
            self.verticalScrollBar().value(),
        ).normalized()
        tl_photo_grid_col_index, tl_photo_grid_row_index = self.photo_grid_position_at(photo_grid_rect.topLeft())
        br_photo_grid_col_index, br_photo_grid_row_index = self.photo_grid_position_at(photo_grid_rect.bottomRight())

        def _paint_rows(row_from, row_to):
            for row in range(row_from, (row_to + 1)):
                index = self.model().index(row, 0, self.rootIndex())
                if not index.isValid():
                    continue
                rect = self.viewport_rect_for_row(row)
                option = QStyleOptionViewItem()
                self.initViewItemOption(option)
                option.rect = rect.toRect()
                if self.selectionModel().isSelected(index):
                    option.state |= QStyle.StateFlag.State_Selected
                if self.currentIndex() == index:
                    option.state |= QStyle.StateFlag.State_HasFocus

                item_delegate = self.itemDelegateForIndex(index)
                if isinstance(item_delegate, PhotoViewItemDelegate):
                    photo = index.data(Qt.ItemDataRole.UserRole)
                    cache_state = photo.get_image(max_size=self.photo_load_max_size, query_cache_state_only=True)
                    pixmap = photo.get_image(max_size=self.photo_load_max_size) if cache_state == CacheState.DATA_READY else None
                    item_delegate.paint_with_pixmap(painter, option, index, pixmap)
                else:
                    item_delegate.paint(painter, option, index)

        row_from = self.row_of_photo_grid_position(tl_photo_grid_col_index, tl_photo_grid_row_index)
        row_to = self.row_of_photo_grid_position(br_photo_grid_col_index, br_photo_grid_row_index)

        num_preload_images = (row_to + 1 - row_from)

        self.pixmap_query_manager.set_max_queries(num_preload_images)
        for row in range(row_from, (row_to + 1)):
            self._preload_photo_image(row)

        extra_preload_images = self.photo_tiles_visible
        row_count = self.model().rowCount()
        num_preload_images = ((row_to + 1 - row_from) + (2 * extra_preload_images))
        self.pixmap_query_manager.set_max_queries(num_preload_images)
        for i in range(1, extra_preload_images):
            row_above = (row_from - i)
            row_below = (row_to + i)
            if row_above > 0:
                self._preload_photo_image(row_above)
            if row_below < row_count:
                self._preload_photo_image(row_below)

        _paint_rows(row_from, row_to)

    def update_view(self):
        self.update_photo_tile_size()
        self.update_rect_photo_grid()
        self.updateGeometries()

    def reset(self):
        super().reset()
        self.update_view()

    def resizeEvent(self, event):
        self.update_view()
        super().resizeEvent(event)

    def updateGeometries(self):
        self.horizontalScrollBar().setSingleStep(self.photo_tile_width)
        self.horizontalScrollBar().setPageStep(self.viewport().width())
        self.horizontalScrollBar().setRange(0, 0)
        self.verticalScrollBar().setSingleStep(self.photo_tile_height / 10)
        self.verticalScrollBar().setPageStep(self.viewport().height())
        self.verticalScrollBar().setRange(0, max(0, (self.rect_photo_grid.bottom() - self.viewport().height())))


class PhotoAssistantPhotoView(PhotoView):
    photos_selected_signal = QSignal(list)
    photo_open_signal = QSignal(object)
    photo_delete_signal = QSignal(object)

    def __init__(self, parent=None):
        super().__init__(parent)
        self.doubleClicked.connect(lambda index: self.photo_open_signal.emit(index.data(Qt.ItemDataRole.UserRole)))

    def selectionChanged(self, selected, deselected):
        photos_selected = [index.data(Qt.ItemDataRole.UserRole) for index in selected.indexes()]
        self.photos_selected_signal.emit(photos_selected)

    def keyPressEvent(self, event):
        if event.key() in [Qt.Key.Key_Enter, Qt.Key.Key_Return]:
            selection_indices = self.selectionModel().selection().indexes()
            if len(selection_indices) > 0:
                self.photo_open_signal.emit(selection_indices[0].data(Qt.ItemDataRole.UserRole))
        elif (
            event.matches(QKeySequence.StandardKey.ZoomIn)
            or (
                # Fix for windows: Key '+' is not mapped properly: event.key() should be 43 but has found to be 93.
                # Resort to nativeVirtualKey() with windows specific codes:
                # https://www.qtcentre.org/threads/68749-keys-map-(cross-platform)
                # https://learn.microsoft.com/en-us/windows/win32/inputdev/virtual-key-codes?redirectedfrom=MSDN
                "PLUS" in QKeySequenceStandardKeyUtils.get_user_string(QKeySequence.StandardKey.ZoomIn)
                and (event.nativeVirtualKey() == 0xbb)
            )
        ):
            self.photo_tiles_per_row -= 1
            return
        elif event.matches(QKeySequence.StandardKey.ZoomOut):
            self.photo_tiles_per_row += 1
            return
        elif event.modifiers() == Qt.KeyboardModifier.ControlModifier:
            if event.key() in (Qt.Key.Key_Delete, Qt.Key.Key_Backspace):
                # ATTENTION, DATA LOSS RISK DUE TO QT IMPLEMENTATION
                # Iteration over selected indices and emitting a delete signal for each
                # element could be implemented like this:
                #
                # selection_indices = self.selectionModel().selection().indexes()
                # for selection_index in selection_indices:
                #     self.photo_delete_signal.emit(selection_index.data(Qt.ItemDataRole.UserRole))
                #
                # But there are some Qt specific limitations/weaknesses that currently
                # and potentially forever will lead to the deletion of the wrong photos!
                #
                # - The selection_indices being fetched once, can change their data after
                # photos are deleted. Probably, the C++ implementation uses an array-like
                # data type and Python gets a reference to a memory that is not bound to
                # the actual data.
                # - A deepcopy does not work as Python can not pickle a QModelIndex.
                # - As the Qt implementation 'removeRows' can only be called for contiguous
                # ranges we cannot achieve the correct behavior within a single instruction
                # (beginRemoveRows, delete_photos_in_the_model, endReomveRows).
                # - Removing photos in reversed order currently works (supports assumption
                # of array-like implementation with reference to memory slots). But this
                # behavior wouldn't be guaranteed to not change in the future!
                # -> The only solution left is handling the consistency promise on our own such
                # that only the photos that are supposed to be deleted will be deleted!
                # -> Fetch all photos upfront in our own list which elements are bound to
                # the data such that we can rely on consistency and work through that list! 

                selection_indices = self.selectionModel().selection().indexes()
                photos_selected = [index.data(Qt.ItemDataRole.UserRole) for index in selection_indices]
                for photo in photos_selected:
                    self.photo_delete_signal.emit(photo)
                return
        super().keyPressEvent(event)



# SPDX-FileCopyrightText: 2023 Anthony Zimmermann
#
# SPDX-License-Identifier: GPL-2.0-only

import os
import re

from photoassistant.core.photo import PhotoAsyncSharedCached
from photoassistant.gui.qt.utils.qtutils import QPixmapLoader
from photoassistant.utils.cachingutils import AsyncSharedCache
from photoassistant.utils.decoratorutils import async_cached_query


class PhotoQt(PhotoAsyncSharedCached):
    ASYNC_SHARED_CACHE_IMAGES = AsyncSharedCache(slots=100)
    SUPPORTED_FILES_RE = re.compile(r"^.*((\.bmp)|(\.gif)|(\.jpe?g)|(\.png)|(\.pbm)|(\.pgm)|(\.ppm)|(\.xbm)|(\.xpm)|(\.svg))$")

    @staticmethod
    def file_is_supported(path):
        return PhotoQt.SUPPORTED_FILES_RE.match(path) is not None and os.path.exists(path)

    #override
    def set_orientation_correction(self, orientation_correction):
        super().set_orientation_correction(orientation_correction)
        self.ASYNC_SHARED_CACHE_IMAGES.delete_cache(self)

    #override
    @async_cached_query(ASYNC_SHARED_CACHE_IMAGES)
    def get_image(self, max_size=None, **kwargs):
        # Special keyword arguments:
        # 'query_ignore_cache':       default=False - set to False if cache should be fully ignored
        # 'query_cache_state_only':   default=False - do not execute the query, only peek into cache if value is available
        # 'query_callback':           default=None - sets a callback to be called when the result is available
        # 'query_result_placeholder': default=AsyncSharedCacheQueryPolicy.NONE - value to be returned if the value is not yet cached
        orientation_correction = self.get_orientation_correction()
        return QPixmapLoader.load(
            self.get_path(),
            max_size=max_size,
            orientation_correction=orientation_correction,
        )



# SPDX-FileCopyrightText: 2023 Anthony Zimmermann
#
# SPDX-License-Identifier: GPL-2.0-only

from photoassistant.utils.cachingutils import AsyncSharedCacheQueryPolicy

class async_cached_query:
    # A decorator that evaluates special keyword arguments and modifies the original function.
    # Enables utilizing AsyncSharedCache, but the cache can also be ignored entirely. Furthermore, the cache can also be inspected.
    # Special keyword arguments:
    # 'query_ignore_cache':       default=False - set to False if cache should be fully ignored
    # 'query_cache_state_only':   default=False - do not execute the query, only peek into cache if value is available
    # 'query_callback':           default=None - sets a callback to be called when the result is available
    # 'query_result_placeholder': default=AsyncSharedCacheQueryPolicy.NONE - value to be returned if the value is not yet cached

    def __init__(self, async_shared_cache):
        self.async_shared_cache = async_shared_cache

    def __call__(self, function):
        def decorated_function(self_, *args, query_ignore_cache=False, query_cache_state_only=False, query_callback=None, query_result_placeholder=AsyncSharedCacheQueryPolicy.NONE, **kwargs):
            if query_ignore_cache:
                return function(self_, *args, **kwargs)

            async_shared_cache_query_policy = AsyncSharedCacheQueryPolicy(
                self_, # self_ is 'self' of the instance calling the decorated method -> here it is also the owner of the query
                function,
                callback=query_callback,
            )
            async_shared_cache_query_policy.set_result_placeholder(query_result_placeholder)

            if query_cache_state_only:
                return self.async_shared_cache.get_cache_state(async_shared_cache_query_policy, self_, *args, **kwargs)
            return self.async_shared_cache.query(async_shared_cache_query_policy, self_, *args, **kwargs)
        return decorated_function



# SPDX-FileCopyrightText: 2023 Anthony Zimmermann
#
# SPDX-License-Identifier: GPL-2.0-only

class AsyncExecutePolicy:
    def __init__(self, function, callback=None):
        self.function = function
        self.callback = callback

    def __call__(self, worker_pool, *args, **kwargs):
        worker_pool.apply_async(
            self.function,
            args,
            kwargs,
            callback=self.callback
        )



# SPDX-FileCopyrightText: 2023 Anthony Zimmermann
#
# SPDX-License-Identifier: GPL-2.0-only

from PySide6.QtCore import Qt
from PySide6.QtCore import QSize
from PySide6.QtGui import QPainter
from PySide6.QtWidgets import QWidget


class ImageViewer(QWidget):
    def __init__(self):
        super().__init__()

        self._pixmap = None
        self._pixmap_scaled = None

        self._aspect_ratio = Qt.AspectRatioMode.KeepAspectRatio
        self._size_hint = QSize()
        self._transformation_mode = Qt.TransformationMode.SmoothTransformation

    def _calc_pixmap_scaled(self):
        if self._pixmap:
            self._pixmap_scaled = self._pixmap.scaled(
                self.size(),
                self._aspect_ratio,
                self._transformation_mode,
            )
        self.update()

    def set_pixmap(self, pixmap):
        if self._pixmap == pixmap:
            return

        self._pixmap = pixmap
        if pixmap is None:
            self._size_hint = QSize(0, 0)
            return

        self._size_hint = pixmap.size()
        self._pixmap_scaled = pixmap.scaled(
            self.size(),
            self._aspect_ratio,
            self._transformation_mode,
        )

        self._calc_pixmap_scaled()
        self.updateGeometry()

    def set_aspect_ratio(self, aspect_ratio):
        if self._aspect_ratio == aspect_ratio:
            return
        self._aspect_ratio = aspect_ratio
        self._calc_pixmap_scaled()

    def set_transformation_mode(self, transformation_mode):
        if self._transformation_mode == transformation_mode:
            return
        self._transformation_mode = transformation_mode
        self._calc_pixmap_scaled()

    def sizeHint(self):
        return self._size_hint

    def resizeEvent(self, event):
        self._calc_pixmap_scaled()
        super().resizeEvent(event)

    def paintEvent(self, event):
        if self._pixmap_scaled is None:
            return
        painter = QPainter(self)
        pixmap_rect = self._pixmap_scaled.rect()
        pixmap_rect.moveCenter(self.rect().center())
        painter.drawPixmap(pixmap_rect, self._pixmap_scaled)
        super().paintEvent(event)



# SPDX-FileCopyrightText: 2023 Anthony Zimmermann
#
# SPDX-License-Identifier: GPL-2.0-only

import datetime
import importlib.resources
import os
import re
import sys
from PySide6.QtCore import Qt
from PySide6.QtCore import Signal as QSignal
from PySide6.QtGui import QAction
from PySide6.QtGui import QIcon
from PySide6.QtGui import QStandardItem
from PySide6.QtGui import QStandardItemModel
from PySide6.QtGui import QKeySequence
from PySide6.QtWidgets import QApplication
from PySide6.QtWidgets import QComboBox
from PySide6.QtWidgets import QFileDialog
from PySide6.QtWidgets import QFrame
from PySide6.QtWidgets import QGridLayout
from PySide6.QtWidgets import QLabel
from PySide6.QtWidgets import QMainWindow
from PySide6.QtWidgets import QMenuBar
from PySide6.QtWidgets import QPushButton
from PySide6.QtWidgets import QScrollArea
from PySide6.QtWidgets import QSizePolicy
from PySide6.QtWidgets import QSpacerItem
from PySide6.QtWidgets import QSplitter
from PySide6.QtWidgets import QStatusBar
from PySide6.QtWidgets import QTabWidget
from PySide6.QtWidgets import QTextEdit
from PySide6.QtWidgets import QWidget

from photoassistant.core.photocollection import PhotoCollectionManager
from photoassistant.gui.qt.photo import PhotoQt
from photoassistant.gui.qt.photodatamodel import PhotoDataModel
from photoassistant.gui.qt.photoview import PhotoAssistantPhotoView
from photoassistant.gui.qt.utils.qtutils import BusyIndicatorDialog
from photoassistant.gui.qt.utils.qtutils import ErrorQMessageBox
from photoassistant.gui.qt.utils.qtutils import InfoQMessageBox
from photoassistant.gui.qt.utils.qtutils import QComboBoxEditable
from photoassistant.gui.qt.utils.qtutils import QKeySequenceStandardKeyUtils
from photoassistant.gui.qt.utils.qtutils import QLabeledProperty
from photoassistant.gui.qt.utils.qtutils import QListWidgetTransparentMinimumSized
from photoassistant.gui.qt.utils.qtutils import QModernStyleGroupBox
from photoassistant.gui.qt.utils.qtutils import QModernStyleGroupBoxContainer
from photoassistant.gui.qt.utils.qtutils import QMarginContainer
from photoassistant.gui.qt.utils.qtutils import QSizePolicyStretch
from photoassistant.gui.qt.utils.qtutils import ScrollableInfoDialog
from photoassistant.gui.qt.utils.qtutils import SynchronizedExecution
from photoassistant.gui.qt.imageviewer import ImageViewer


class TagsModel(QStandardItemModel):
    def __init__(self):
        super().__init__()

    def set_tags(self, tags):
        self.clear()
        self.appendRow(QStandardItem(""))
        for tag in tags:
            self.appendRow(QStandardItem(tag))


class PhotoViewControlPanel(QWidget):
    sort_function_selected_signal = QSignal(str)
    apply_filters_signal = QSignal(list, list)

    def __init__(self, parent=None, **kwargs):
        super().__init__(parent, **kwargs)

        sort_combobox = QComboBox()
        sort_combobox.addItems(
            [
                "filename",
                "creation timestamp"
            ]
        )

        self.filter_include_text_input = QTextEdit()
        self.filter_include_text_input.setMinimumWidth(100)
        self.filter_include_text_input.setPlaceholderText(
            f"Every line adds a filter rule. "
            f"The following two lines filter photos that either have a tag including '2023' or have the tag 'city':{os.linesep}"
            f"*2023*{os.linesep}"
            f"city{os.linesep}"
        )
        self.filter_exclude_text_input = QTextEdit()
        self.filter_exclude_text_input.setMinimumWidth(100)
        self.filter_exclude_text_input.setPlaceholderText(
            f"Every line adds a filter rule. "
            f"The following two lines filter photos that do not have a tag including '2023' and do not have the tag 'city':{os.linesep}"
            f"*2023*{os.linesep}"
            f"city{os.linesep}"
        )

        apply_filters_button = QPushButton("Apply Tag Filters")
        apply_filters_button.clicked.connect(self._emit_apply_filters)

        apply_filters_button_layout = QGridLayout()
        apply_filters_button_layout.addWidget(apply_filters_button, 0, 1)
        apply_filters_button_layout.setColumnStretch(0, 1)
        apply_filters_button_layout.setColumnStretch(1, 18)
        apply_filters_button_layout.setColumnStretch(2, 1)


        sort_combobox.currentTextChanged.connect(self.sort_function_selected_signal.emit)

        grid_layout = QGridLayout(self)
        grid_layout.setContentsMargins(0, 0, 0, 0)
        grid_layout.addWidget(QModernStyleGroupBoxContainer("Sort By", widget=sort_combobox), 0, 0)
        grid_layout.addWidget(QModernStyleGroupBoxContainer("Filter Photos With Tags", widget=self.filter_include_text_input), 1, 0)
        grid_layout.addWidget(QModernStyleGroupBoxContainer("Filter Photos Without Tags", widget=self.filter_exclude_text_input), 2, 0)
        grid_layout.addLayout(apply_filters_button_layout, 3, 0)

    def _emit_apply_filters(self):
        include_filter_tags = list()
        include_filter_text = self.filter_include_text_input.toPlainText()
        for line in include_filter_text.splitlines():
            line = line.strip()
            include_filter_tags.append(rf"^{line.replace('*', '.*')}$")

        exclude_filter_tags = list()
        exclude_filter_text = self.filter_exclude_text_input.toPlainText()
        for line in exclude_filter_text.splitlines():
            line = line.strip()
            exclude_filter_tags.append(rf"^{line.replace('*', '.*')}$")
        self.apply_filters_signal.emit(include_filter_tags, exclude_filter_tags)


class TagEntry(QWidget):
    def __init__(self, parent=None, **kwargs):
        super().__init__(parent, **kwargs)
        self.setMinimumHeight(20)

        self.tag_entry = QComboBoxEditable(minimum_size=(100, 0))

        grid_layout = QGridLayout(self)
        grid_layout.setContentsMargins(0, 0, 0, 0)
        grid_layout.addWidget(self.tag_entry, 0, 0)

    def set_model(self, model):
        if model is None:
            tag_entry_model = self.tag_entry.model()
            if isinstance(tag_entry_model, TagsModel):
                tag_entry_model.clear()
            return
        self.tag_entry.setModel(model)

    def _set_tag_items(self, tags):
        assert isinstance(tags, list)
        self.tag_entry.addItems(tags)

    def text(self):
        return self.tag_entry.currentText()


class TagGroup(QWidget):
    delete_signal = QSignal()

    class TagEntryWithDeleteButton(TagEntry):
        def __init__(self, parent=None, **kwargs):
            super().__init__(parent, **kwargs)
            self.tag_delete_button = QPushButton("X")
            self.tag_delete_button.setMinimumWidth(20)
            self.tag_delete_button.setMaximumWidth(40)

    def __init__(self, parent=None, **kwargs):
        super().__init__(parent, **kwargs)
        self.tags_model = None

        self.tag_entries = []
        self.grid_layout_tag_entries = QGridLayout()
        self.grid_layout_tag_entries.setContentsMargins(0, 0, 0, 0)

        add_tag_button = QPushButton("+")

        add_tag_button_layout = QGridLayout()
        add_tag_button_layout.addWidget(add_tag_button, 0, 1)
        add_tag_button_layout.setColumnStretch(0, 1)
        add_tag_button_layout.setColumnStretch(1, 1)
        add_tag_button_layout.setColumnStretch(2, 1)

        add_tag_button.clicked.connect(self.add_tag_entry)

        self.grid_layout = QGridLayout(self)
        self.grid_layout.setContentsMargins(0, 0, 0, 0)
        self.grid_layout.addLayout(self.grid_layout_tag_entries, 0, 0)
        self.grid_layout.addLayout(add_tag_button_layout, 1, 0)

        self.add_tag_entry()

    def set_model(self, model):
        self.tags_model = model
        for tag_entry in self.tag_entries:
            tag_entry.set_model(model)

    def add_tag_entry(self):
        tag_entry = self.TagEntryWithDeleteButton()
        tag_entry.set_model(self.tags_model)
        tag_index = len(self.tag_entries)
        self.tag_entries.append(tag_entry)
        self.grid_layout_tag_entries.addWidget(tag_entry, tag_index, 0)
        self.grid_layout_tag_entries.addWidget(tag_entry.tag_delete_button, tag_index, 1)

        tag_entry.tag_delete_button.clicked.connect(lambda: self.delete_tag_entry(tag_entry))

    def delete_tag_entry(self, tag_entry):
        tag_index = self.tag_entries.index(tag_entry)
        for i in reversed(range(tag_index, len(self.tag_entries))):
            tag_entry_widget = self.tag_entries[i]
            tag_entry_widget.setParent(None)
            tag_delete_button_widget = tag_entry_widget.tag_delete_button
            tag_delete_button_widget.setParent(None)
        self.tag_entries.pop(tag_index).deleteLater()
        for i, tag_entry in enumerate(self.tag_entries[tag_index:], tag_index):
            self.grid_layout_tag_entries.addWidget(tag_entry, i, 0)
            self.grid_layout_tag_entries.addWidget(tag_entry.tag_delete_button, i, 1)

        if len(self.tag_entries) == 0:
            self.delete_signal.emit()

    def get_tags(self):
        all_tags = []
        for tag_entry in self.tag_entries:
            tag_text = tag_entry.text().strip()
            if tag_text != "":
                all_tags.append(tag_text)
        return all_tags


class TagsPanel(QWidget):
    tags_updated_signal = QSignal()

    class TagGroupWithHeader(QModernStyleGroupBox):
        delete_signal = QSignal()

        def __init__(self, title="Tag Group X", parent=None):
            super().__init__(title, parent)

            self.apply_tag_group_button = QPushButton(f"Apply Tags")
            self.tag_group = TagGroup()

            grid_layout = QGridLayout(self)
            grid_layout.setContentsMargins(0, 0, 0, 0)
            grid_layout.addWidget(self.apply_tag_group_button, 0, 0)
            grid_layout.addWidget(self.tag_group, 1, 0)

            self.tag_group.delete_signal.connect(self.delete_signal.emit)

        def set_model(self, model):
            self.tag_group.set_model(model)

        def set_tag_group_index(self, index):
            self.setTitle(f"Tag Group {index}")
            self.apply_tag_group_button.setText(f"Apply Tags")

        def get_tags(self):
            return self.tag_group.get_tags()

    def __init__(self, parent=None, **kwargs):
        super().__init__(parent, **kwargs)
        self.tags_model = None

        self.tag_groups = []
        self.photos_selected = []

        self.grid_layout_tag_groups = QGridLayout()
        self.grid_layout_tag_groups.setContentsMargins(0, 0, 0, 0)

        add_tag_group_button = QPushButton("Add Tag Group")

        add_tag_group_button_layout = QGridLayout()
        add_tag_group_button_layout.addWidget(add_tag_group_button, 0, 1)
        add_tag_group_button_layout.setColumnStretch(0, 1)
        add_tag_group_button_layout.setColumnStretch(1, 18)
        add_tag_group_button_layout.setColumnStretch(2, 1)

        add_tag_group_button.clicked.connect(self.add_tag_group)

        grid_layout = QGridLayout(self)
        grid_layout.setContentsMargins(0, 0, 0, 0)
        grid_layout.addLayout(self.grid_layout_tag_groups, 0, 0)
        grid_layout.addLayout(add_tag_group_button_layout, 1, 0)
        grid_layout.addItem(QSpacerItem(0, 1, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding), 2, 0)

        self.add_tag_group()

    def set_model(self, model):
        self.tags_model = model
        for tag_group in self.tag_groups:
            tag_group.set_model(model)

    def add_tag_group(self):
        tag_group = TagsPanel.TagGroupWithHeader()
        tag_group.set_model(self.tags_model)
        tag_group_index = len(self.tag_groups)
        tag_group.set_tag_group_index((tag_group_index + 1))
        self.tag_groups.append(tag_group)
        self.grid_layout_tag_groups.addWidget(tag_group, tag_group_index, 0)

        tag_group.delete_signal.connect(lambda: self.delete_tag_group(tag_group))
        tag_group.apply_tag_group_button.clicked.connect(lambda: self.apply_tags(tag_group_index))

    def delete_tag_group(self, tag_group):
        tag_group_index = self.tag_groups.index(tag_group)
        for i in reversed(range(tag_group_index, len(self.tag_groups))):
            tag_group = self.tag_groups[i]
            tag_group.setParent(None)
        self.tag_groups.pop(tag_group_index).deleteLater()
        for i, tag_group in enumerate(self.tag_groups[tag_group_index:], tag_group_index):
            self.grid_layout_tag_groups.addWidget(tag_group, i, 0)
            tag_group.set_tag_group_index((i + 1))

    def set_photos_selected(self, photos):
        self.photos_selected = photos

    def apply_tags(self, tag_group_index):
        if tag_group_index < 0 or tag_group_index >= len(self.tag_groups):
            return
        tags = self.tag_groups[tag_group_index].get_tags()
        main_tag = [tag for tag in tags if tag.startswith(":")]
        main_tag = None if len(main_tag) == 0 else main_tag[-1]
        tags = [tag for tag in tags if not tag.startswith(":")]
        for photo in self.photos_selected:
            if main_tag is not None:
                photo.set_main_tag(main_tag)
            for tag in tags:
                photo.add_tag(tag)
        self.tags_updated_signal.emit()


class PhotoInfoPanel(QWidget):
    class SinglePhotoInfo(QWidget):
        def __init__(self, parent=None, **kwargs):
            super().__init__(parent, **kwargs)

            self.photo = None

            self.directory_property = QLabeledProperty("File Location")
            self.filename_property = QLabeledProperty("Filename")
            self.creation_timestamp_property = QLabeledProperty("Creation Date")
            self.gps_coordinates_property = QLabeledProperty("GPS Coordinates")

            grid_layout = QGridLayout(self)
            grid_layout.setContentsMargins(0, 0, 0, 0)
            grid_layout.addWidget(self.directory_property, 0, 0)
            grid_layout.addWidget(self.filename_property, 1, 0)
            grid_layout.addWidget(self.creation_timestamp_property, 2, 0)
            grid_layout.addWidget(self.gps_coordinates_property, 3, 0)

            self.set_photo(None)

        def set_photo(self, photo):
            if photo is None:
                self.hide()
                return

            self.show()
            self.photo = photo
            self.directory_property.set_text(f"{os.path.dirname(photo.path)}")
            self.filename_property.set_text(f"{photo.filename}")
            self.creation_timestamp_property.set_text(f"{photo.creation_timestamp}")
            self.gps_coordinates_property.set_text(f"{' '.join(f'{val:.4f}' if val is not None else '?' for val in photo.gps_coordinates)}")

    def __init__(self, parent=None, **kwargs):
        super().__init__(parent, **kwargs)

        self.photos_selected = []

        self.single_photo_info = self.SinglePhotoInfo()
        self.tag_list_widget = QListWidgetTransparentMinimumSized(self)
        self.tag_list_widget.setMinimumSize(100, 50)
        tag_list_widget_container = QModernStyleGroupBoxContainer("Tags", widget=self.tag_list_widget)

        grid_layout = QGridLayout(self)
        grid_layout.setContentsMargins(0, 0, 0, 0)
        grid_layout.addWidget(self.single_photo_info, 0, 0)
        grid_layout.addWidget(tag_list_widget_container, 1, 0)

    def set_photos_selected(self, photos):
        self.photos_selected = photos
        self.update_photo_info()

    def update_tag_list(self):
        self.tag_list_widget.clear()
        tags = set()
        for photo in self.photos_selected:
            tags.update(photo.get_all_tags())
        main_tags = [tag for tag in tags if tag.startswith(":")]
        tags = [tag for tag in tags if not tag.startswith(":")]
        for tag in sorted(main_tags):
            self.tag_list_widget.addItem(tag)
        for tag in sorted(tags):
            self.tag_list_widget.addItem(tag)

    def update_photo_info(self):
        self.single_photo_info.set_photo(self.photos_selected[0] if len(self.photos_selected) == 1 else None)
        self.update_tag_list()

    def keyPressEvent(self, event):
        if event.key() in (Qt.Key.Key_Delete, Qt.Key.Key_Backspace):
            selected_tags = [item.text() for item in self.tag_list_widget.selectedItems()]
            main_tags = [tag for tag in selected_tags if tag.startswith(":")]
            tags = [tag for tag in selected_tags if not tag.startswith(":")]
            for photo in self.photos_selected:
                if photo.main_tag in main_tags:
                    photo.main_tag = None
                for tag in tags:
                    photo.delete_tag(tag)
            self.update_photo_info()
            return
        super().keyPressEvent(event)


class PhotoViewer(QWidget):
    def __init__(self, parent=None, **kwargs):
        super().__init__(parent, **kwargs)

        self.photo = None

        self.image_viewer = ImageViewer()

        grid_layout = QGridLayout(self)
        grid_layout.setContentsMargins(0, 0, 0, 0)
        grid_layout.addWidget(self.image_viewer, 0, 0)

    def update_photo(self, photo):
        self.photo = photo
        self.image_viewer.set_pixmap(photo.get_image(max_size=self.image_viewer.size(), query_ignore_cache=True))

    def set_photo(self, photo):
        self.update_photo(photo)
        self.show()

    def resizeEvent(self, event):
        super().resizeEvent(event)
        self.update_photo(self.photo)


class PhotoSortFunctions:
    DATETIME_NONE = datetime.datetime.fromtimestamp(0)
    @staticmethod
    def filename(photo):
        return photo.get_filename()

    @staticmethod
    def creation_timestamp(photo):
        timestamp = photo.get_creation_timestamp()
        if timestamp is None:
            timestamp = PhotoSortFunctions.DATETIME_NONE
        return timestamp

class PhotoAssistant(QMainWindow):
    apply_tags_signal = QSignal(int)
    photo_updated_signal = QSignal(object)
    PHOTOS_SORT_FUNCTIONS = {
        "filename": PhotoSortFunctions.filename,
        "creation timestamp": PhotoSortFunctions.creation_timestamp,
    }
    def __init__(self, parent=None, **kwargs):
        super().__init__(parent, **kwargs)

        with importlib.resources.path("photoassistant", "photoassistant-icon.png") as icon_path:
            self.icon_path = str(icon_path.resolve())
        icon = QIcon(self.icon_path)

        self.setWindowTitle(self.__class__.__name__)
        self.setWindowIcon(icon)
        self.setMinimumSize(640, 480)

        # business logic initialization
        self.photo_collection = PhotoCollectionManager(PhotoQt)
        self.photo_index = None

        self.photo_data_model_qt = PhotoDataModel()
        self.tags_model_qt = TagsModel()

        self.photos_sort_key = "filename"
        self.photo_filters = ([], [])
        self.photos_selected = []

        # Special widgets:
        # - seperate photo viewer window
        # - permanent status bar widget
        # - busy indicator dialog
        photo_viewer = PhotoViewer()
        self.programm_status_info_label = QLabel()
        self.busy_indicator_dialog = BusyIndicatorDialog(self)
        self.busy_indicator_dialog.setMinimumWidth(self.minimumWidth())
        self.busy_indicator_dialog.setWindowIcon(icon)
        self.info_dialog = ScrollableInfoDialog()
        self.info_dialog.setMinimumSize(self.minimumSize())
        self.info_dialog.setWindowIcon(icon)

        # central widget 'PhotoView' (grid with photos)
        photo_view = PhotoAssistantPhotoView()
        photo_view.setModel(self.photo_data_model_qt)
        photo_view.setSizePolicy(QSizePolicyStretch(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred, stretch_horizontal=3, stretch_vertical=1))

        # side panel widgets
        photo_view_control_panel = PhotoViewControlPanel()

        photo_add_tags_panel = TagsPanel()
        photo_add_tags_panel.set_model(self.tags_model_qt)

        photo_add_tags_panel_scroll_container = QScrollArea()
        photo_add_tags_panel_scroll_container.setWidgetResizable(True)
        photo_add_tags_panel_scroll_container.setFrameShape(QFrame.Shape.NoFrame)
        photo_add_tags_panel_scroll_container.setWidget(photo_add_tags_panel)

        info_panel = PhotoInfoPanel()

        photo_control_panel = QTabWidget()
        photo_control_panel.setDocumentMode(True)

        photo_control_panel.addTab(photo_add_tags_panel_scroll_container, "Add Tags")
        photo_control_panel.addTab(photo_view_control_panel, "View Control")

        # side panel widgets go into a vertical splitter
        side_panel_splitter = QSplitter()
        side_panel_splitter.setHandleWidth(8)
        side_panel_splitter.setOrientation(Qt.Orientation.Vertical)
        side_panel_splitter.addWidget(photo_control_panel)
        side_panel_splitter.addWidget(info_panel)
        side_panel_splitter.setSizes([800, 200])

        # central widget and side panel go into a horizontal splitter
        main_splitter = QSplitter()
        main_splitter.setHandleWidth(8)
        main_splitter.setOrientation(Qt.Orientation.Horizontal)
        main_splitter.addWidget(photo_view)
        main_splitter.addWidget(side_panel_splitter)
        main_splitter.setSizes([1800, 200])

        # signals and slots
        photo_view_control_panel.sort_function_selected_signal.connect(self.sort_photos)
        photo_view_control_panel.apply_filters_signal.connect(self.filter_photos)
        photo_view.photos_selected_signal.connect(info_panel.set_photos_selected)
        photo_view.photos_selected_signal.connect(lambda photos: photo_viewer.update_photo(photos[0]) if len(photos) == 1 else None)
        photo_view.photos_selected_signal.connect(self.set_photos_selected)
        photo_view.photos_selected_signal.connect(photo_add_tags_panel.set_photos_selected)
        photo_view.photo_open_signal.connect(photo_viewer.set_photo)
        photo_view.photo_delete_signal.connect(self.delete_photo)
        photo_add_tags_panel.tags_updated_signal.connect(info_panel.update_photo_info)
        self.apply_tags_signal.connect(photo_add_tags_panel.apply_tags)
        self.photo_updated_signal.connect(photo_viewer.update_photo)

        self.setCentralWidget(QMarginContainer(widget=main_splitter, margin=(8, 8, 8, 8)))

        self._initialize_menu_bar()


    def keyPressEvent(self, event):
        if event.modifiers() == Qt.KeyboardModifier.ControlModifier:
            number_keys = (Qt.Key.Key_1, Qt.Key.Key_2, Qt.Key.Key_3, Qt.Key.Key_4, Qt.Key.Key_5, Qt.Key.Key_6, Qt.Key.Key_7, Qt.Key.Key_8, Qt.Key.Key_9)
            if event.key() in number_keys: 
                index = number_keys.index(event.key())
                self.apply_tags_signal.emit(index)
                return
            if event.key() == Qt.Key.Key_R:
                if len(self.photos_selected) == 1:
                    self.rotate_photo(self.photos_selected[0])
                return
        super().keyPressEvent(event)

    def closeEvent(self, event):
        QApplication.quit()

    def _initialize_menu_bar(self):
        create_photo_collection_action = QAction("Create Collection", self)
        create_photo_collection_action.setStatusTip("Create a new Collection.")
        open_photo_collection_action = QAction("Open &Collection", self)
        open_photo_collection_action.setStatusTip("Open an existing Collection.")
        open_photos_action = QAction("Open &Photos", self)
        open_photos_action.setStatusTip("Open photos.")
        integrate_opened_photos_action = QAction("Integrate Opened Photos and Meta Data", self)
        integrate_opened_photos_action.setStatusTip("Integrate currently opened photos and their meta data.")

        open_settings_action = QAction("Settings", self)
        open_settings_action.setStatusTip("Get help on changing settings of PhotoAssistant.")

        show_photoassistant_introduction_action = QAction("PhotoAssistant Introduction", self)
        show_photoassistant_introduction_action.setStatusTip("Displays an introduction to PhotoAssistant.")

        show_getting_started_action = QAction("Getting Started", self)
        show_getting_started_action.setStatusTip("Display a \"Getting Started\" help page.")

        show_quick_reference_action = QAction("Quick Reference", self)
        show_quick_reference_action.setStatusTip("Display a \"Quick Reference\" page.")

        show_about_action = QAction("About", self)
        show_about_action.setStatusTip("Display an \"About\" page.")


        create_photo_collection_action.triggered.connect(self._choose_photo_collection_to_create)
        open_photo_collection_action.triggered.connect(self._choose_photo_collection_to_open)
        open_photos_action.triggered.connect(self._choose_photos_to_open)
        integrate_opened_photos_action.triggered.connect(self._integrate_opened_photos)
        open_settings_action.triggered.connect(self._open_settings)

        show_photoassistant_introduction_action.triggered.connect(self._display_photoassistant_introduction)
        show_getting_started_action.triggered.connect(self._display_getting_started)
        show_quick_reference_action.triggered.connect(self._display_quick_reference)
        show_about_action.triggered.connect(self._display_about)

        menu = QMenuBar(None)
        self.setMenuBar(menu)

        file_menu = menu.addMenu("&File")
        file_menu.addAction(create_photo_collection_action)
        file_menu.addAction(open_photo_collection_action)
        file_menu.addAction(open_photos_action)
        file_menu.addAction(integrate_opened_photos_action)
        file_menu.addAction(open_settings_action)

        help_menu = menu.addMenu("&Help")
        help_menu.addAction(show_photoassistant_introduction_action)
        help_menu.addAction(show_getting_started_action)
        help_menu.addAction(show_quick_reference_action)
        help_menu.addAction(show_about_action)

        status_bar = QStatusBar(self)
        status_bar.addPermanentWidget(self.programm_status_info_label)

        self.setStatusBar(status_bar)

    def set_photos_selected(self, photos):
        self.photos_selected = photos

    def create_photo_collection(self, directory):
        try:
            collection_root = PhotoCollectionManager.find_collection(directory)
            assert collection_root is None, f"Collections can not be nested. The conflicting collection is '{collection_root}'."
        except PhotoCollectionManager.CollectionError as e:
            return
        except AssertionError as e:
            return

        PhotoCollectionManager.create(directory)

    def open_photo_collection(self, directory):
        # This function must be executed in a thread seperate from the main thread
        # -> Using self.busy_indicator_dialog.execute 

        def _update_ui():
            self.set_program_status_info_label_text()

        update_ui_event = SynchronizedExecution.execute_event_blocked(_update_ui)
        def _update_data():
            try:
                collection_root = PhotoCollectionManager.find_collection(directory)
                assert collection_root is not None, f"Collection can not be found at '{directory}'. Please use the menu to create a collection first!"
            except PhotoCollectionManager.CollectionError as e:
                return  # decorator will return BusyIndicatorDialog.ExecuteResult(None)
            except AssertionError:
                return  # decorator will return BusyIndicatorDialog.ExecuteResult(None)

            self.photo_collection.set_collection_root(directory)
            update_ui_event.set()

        self.busy_indicator_dialog.execute("Open photo collection.", _update_data)

    def _update_photo_index(self):
        # This function must be executed in a thread seperate from the main thread
        # -> Using self.busy_indicator_dialog.execute 

        def _update_data():
            if self.photo_collection.collection_root is None or self.photo_collection.photo_index is None:
                return  # decorator will return BusyIndicatorDialog.ExecuteResult(None)

            photo_index_copy = self.photo_collection.photo_index.copy()
            include_tag_patterns, exclude_tag_patterns = self.photo_filters

            def _tags_include_match(photo):
                if len(include_tag_patterns) == 0:
                    return True
                photo_tags = photo.get_all_tags()
                for include_tag_pattern in include_tag_patterns:
                    if any(re.match(include_tag_pattern, tag) for tag in photo_tags):
                        return True
                return False

            def _tags_exclude_match(photo):
                if len(exclude_tag_patterns) == 0:
                    return False
                photo_tags = photo.get_all_tags()
                for exclude_tag_pattern in exclude_tag_patterns:
                    if any(re.match(exclude_tag_pattern, tag) for tag in photo_tags):
                        return True
                return False

            if len(include_tag_patterns) > 0:
                photo_index_copy = self.busy_indicator_dialog.execute(
                    "Filter photos by list of include tags.",
                    lambda: [photo for photo in photo_index_copy if _tags_include_match(photo)],
                ).wait()

            if len(exclude_tag_patterns) > 0:
                photo_index_copy = self.busy_indicator_dialog.execute(
                    "Filter photos by list of exclude tags.",
                    lambda: [photo for photo in photo_index_copy if not _tags_exclude_match(photo)],
                ).wait()

            self.busy_indicator_dialog.execute(
                f"Sort photos by '{self.photos_sort_key}'. This might require reading the meta data of every photo and meta file once.",
                lambda: photo_index_copy.sort(key=self.PHOTOS_SORT_FUNCTIONS[self.photos_sort_key]),
            ).wait()

            self.photo_index = photo_index_copy

        self.busy_indicator_dialog.execute("Update opened photos.", _update_data).wait()

    def open_photos(self, path):
        # This function must be executed in a thread seperate from the main thread
        # -> Using self.busy_indicator_dialog.execute 

        results_container = dict()
        def _update_ui():
            self.tags_model_qt.set_tags(results_container["all_tags"])
            self.photo_data_model_qt.set_photo_index(self.photo_index)
            self.set_program_status_info_label_text()

        update_ui_event = SynchronizedExecution.execute_event_blocked(_update_ui)
        def _update_data():
            try:
                assert self.photo_collection.collection_root is not None, "Photos can only be opened after a Collection is opened. Please open a Collection first!"
                assert os.path.isdir(path), f"The path {path} does not exist. Please choose an existing path to open photos!"
            except AssertionError as e:
                return  # decorator will return BusyIndicatorDialog.ExecuteResult(None)

            num_directories = self.busy_indicator_dialog.execute(
                f"Prescan directory '{path}'.",
                lambda: len([elem for elem in os.walk(path)]),
            ).wait()
            self.busy_indicator_dialog.execute(
                f"Load photos from {num_directories} directories in '{path}'.",
                lambda: self.photo_collection.open_photos(path, sort_key=self.PHOTOS_SORT_FUNCTIONS["filename"])
            ).wait()
            num_photos = len(self.photo_collection.photo_index)
            results_container["all_tags"] = self.busy_indicator_dialog.execute(
                f"Read tags from {num_photos} photos.",
                self.photo_collection.get_list_of_all_tags,
            ).wait()
            self._update_photo_index()
            update_ui_event.set()

        self.busy_indicator_dialog.execute("Open photos.", _update_data)

    def sort_photos(self, sort_key="filename"):
        # This function must be executed in a thread seperate from the main thread
        # -> Using self.busy_indicator_dialog.execute 

        def _update_ui():
            self.photo_data_model_qt.set_photo_index(self.photo_index)
            self.set_program_status_info_label_text()

        update_ui_event = SynchronizedExecution.execute_event_blocked(_update_ui)
        def _update_data():
            assert sort_key in self.PHOTOS_SORT_FUNCTIONS
            self.photos_sort_key = sort_key
            self._update_photo_index()
            update_ui_event.set()

        self.busy_indicator_dialog.execute("Sort opened photos.", _update_data)

    def filter_photos(self, include_filter_tags, exclude_filter_tags):
        # This function must be executed in a thread seperate from the main thread
        # -> Using self.busy_indicator_dialog.execute 

        def _update_ui():
            self.photo_data_model_qt.set_photo_index(self.photo_index)
            self.set_program_status_info_label_text()

        update_ui_event = SynchronizedExecution.execute_event_blocked(_update_ui)
        def _update_data():
            self.photo_filters = (include_filter_tags, exclude_filter_tags)
            self._update_photo_index()
            update_ui_event.set()

        self.busy_indicator_dialog.execute("Filter opened photos.", _update_data)

    def rotate_photo(self, photo):
        with self.photo_data_model_qt.updating_photo(photo):
            orientation_correction = photo.get_orientation_correction()
            orientation_correction = [
                orientation_correction[3],
                orientation_correction[4],
                orientation_correction[5],
                -orientation_correction[0],
                -orientation_correction[1],
                -orientation_correction[2],
                orientation_correction[6],
                orientation_correction[7],
                orientation_correction[8],
            ]
            photo.set_orientation_correction(orientation_correction)
        self.photo_updated_signal.emit(photo)

    def delete_photo(self, photo):
        with self.photo_data_model_qt.removing_photo(photo):
            self.photo_collection.delete_photo(photo)
            # photo_data_model_qt.photo_index is a copy of photo_collection.photo_index
            # -> we must delete the photo here seperately.
            self.photo_data_model_qt.photo_index.remove(photo)

    def set_program_status_info_label_text(self):
        collection_root_str = f"Collection: '{self.photo_collection.collection_root}'"
        photos_str = "Photos: no photos loaded"
        if self.photo_collection.photo_index is not None:
            photos_root = self.photo_collection.photo_index.root_path
            photos_str = f"Photos: '{photos_root}'"
            if self.photo_index is not None:
                num_photos = len(self.photo_index)
                photos_str = f"Photos: '{photos_root}' ({num_photos} photos)"
        self.programm_status_info_label.setText(f"{collection_root_str} | {photos_str}")

    def _choose_photo_collection_to_create(self):
        directory = QFileDialog.getExistingDirectory(self, "Create a Collection:")
        if directory is None or directory == "":
            # cancelled operation
            return

        try:
            collection_root = PhotoCollectionManager.find_collection(directory)
            assert collection_root is None, f"Collections can not be nested. The conflicting collection is '{collection_root}'. More information can be found in menu \"Help\""
        except PhotoCollectionManager.CollectionError as e:
            ErrorQMessageBox(self, f"Error encountered during checks for collection creation: {e}").exec()
            return
        except AssertionError as e:
            ErrorQMessageBox(self, str(e)).exec()
            return

        self.create_photo_collection(directory)

    def _choose_photo_collection_to_open(self):
        directory = QFileDialog.getExistingDirectory(self, "Open a Collection:")
        if directory is None or directory == "":
            # cancelled operation
            return

        try:
            collection_root = PhotoCollectionManager.find_collection(directory)
            assert collection_root is not None, f"Collection can not be found at '{directory}'. Please use the menu \"File\" to create a collection first! More information can be found in menu \"Help\""
        except PhotoCollectionManager.CollectionError as e:
            ErrorQMessageBox(self, f"Error encountered when trying to opening a collection at '{directory}': {e}").exec()
            return 
        except AssertionError as e:
            ErrorQMessageBox(self, str(e)).exec()
            return

        if directory != collection_root:
            InfoQMessageBox(self, f"The directory chosen is contained within a collection at '{collection_root}' which will be loaded now.").exec()

        self.open_photo_collection(directory)

    def _choose_photos_to_open(self):
        directory = QFileDialog.getExistingDirectory(self, "Open Photos:")
        if directory is None or directory == "":
            # cancelled operation
            return

        try:
            assert self.photo_collection.collection_root is not None, "Photos can only be opened after a Collection is opened. Please open a Collection first! More information can be found in menu \"Help\""
            assert os.path.isdir(directory), f"The path {directory} does not exist. Please choose an existing path to open photos!"
        except AssertionError as e:
            ErrorQMessageBox(self, str(e)).exec()
            return

        self.open_photos(directory)

    def _integrate_opened_photos(self):
        text = (
            f"This operation will:{os.linesep}"
            f"- delete duplicate photos within the opened photos{os.linesep}"
            f"- cleanup meta data which is stored among the opened photos{os.linesep}"
            f"- integrate opened photos into the currently opened collection{os.linesep}"
            f"{os.linesep}"
            f"The operation will likely take a little while.{os.linesep}"
        )
        info = InfoQMessageBox(self, text)
        info.setStandardButtons(info.StandardButton.Ok | info.StandardButton.Cancel)
        if info.exec() != info.StandardButton.Ok:
            return

        def _update_ui():
            self.photo_data_model_qt.set_photo_index(self.photo_index)
            self.set_program_status_info_label_text()

        update_ui_event = SynchronizedExecution.execute_event_blocked(_update_ui)
        def _update_data():
            num_photos = len(self.photo_collection.photo_index)
            self.busy_indicator_dialog.execute(
                f"Analyze {num_photos} photos for integration.",
                lambda: self.photo_collection.cleanup_and_integrate(),
            ).wait()
            self._update_photo_index()
            update_ui_event.set()

        self.busy_indicator_dialog.execute("Integrate and cleanup opened photos.", _update_data)

    def _get_readme_text(self, mark=None):
        text = ""
        with importlib.resources.path("photoassistant", "README.md") as readme_path:
            with open(readme_path.resolve()) as file_:
                text = file_.read()
        if mark is not None:
            text_match = re.search(rf"<!-- {mark}_START -->(.*)<!-- {mark}_END -->", text, re.DOTALL)
            if text_match is not None:
                text = text_match.group(1)

            text = re.sub(r"\[(.*?)\]\((http.*?)\)", lambda match: f"**{match.group(1)}** ({match.group(2)})", text)
            text = re.sub(r"!\[Image\]\(photoassistant/photoassistant-icon.png\)", lambda _: f"![Image]({self.icon_path})", text)
            for q_standard_key in [
                QKeySequence.StandardKey.ZoomIn,
                QKeySequence.StandardKey.ZoomOut,
            ]:

                key_tag = f"QKeySequence.StandardKey.{q_standard_key.name}"
                text = re.sub(rf"<!-- {key_tag}_START -->.*<!-- {key_tag}_END -->", lambda _: f"`{QKeySequenceStandardKeyUtils.get_user_string(q_standard_key)}`", text)
        return text.strip()

    def _open_settings(self):
        text = self._get_readme_text("SETTINGS_SECTION")
        self.info_dialog.set_markdown(text, title="PhotoAssistant - Settings")

    def _display_photoassistant_introduction(self):
        text = self._get_readme_text("PHOTOASSISTANT_INTRODUCTION_SECTION")
        self.info_dialog.set_markdown(text, title="PhotoAssistant - Introduction")

    def _display_getting_started(self):
        text = self._get_readme_text("GETTING_STARTED_SECTION")
        self.info_dialog.set_markdown(text, title="PhotoAssistant - Getting Started")

    def _display_quick_reference(self):
        text = self._get_readme_text("QUICK_REFERENCE_SECTION")
        self.info_dialog.set_markdown(text, title="PhotoAssistant - Quick Reference")

    def _display_about(self):
        text = self._get_readme_text("ABOUT_SECTION")
        self.info_dialog.set_markdown(text, title="PhotoAssistant - About")




def main():
    app = QApplication(sys.argv)

    photoassistant = PhotoAssistant()
    photoassistant.resize(photoassistant.minimumSize())
    photoassistant.show()

    return app.exec()

if __name__ == "__main__":
    sys.exit(main())



# SPDX-FileCopyrightText: 2023 Anthony Zimmermann
#
# SPDX-License-Identifier: GPL-2.0-only

import logging
import datetime
import os
import re
import shutil

from photoassistant.core.filescan import FileSystemScanner
from photoassistant.core.metastorage import MetaStorage
from photoassistant.utils.descriptors import SettingsDescriptor

logger = logging.getLogger(__file__)
logger.setLevel(logging.INFO)


class PhotoIndex:
    def __init__(self, photo_delegate):
        self.root_path = None
        self._photos = []
        self._photo_delegate = photo_delegate

    def load(self, path, sort_key=None):
        self.root_path = path
        self._photos = self.load_photos()
        if sort_key is not None:
            self.sort(sort_key)

    def copy(self):
        photo_index = PhotoIndex(self._photo_delegate)
        photo_index.root_path = self.root_path
        photo_index._photos = self._photos.copy()
        return photo_index

    def sort(self, key):
        self._photos.sort(key=key)

    def __getitem__(self, index):
        return self._photos[index]

    def __len__(self):
        return len(self._photos)

    def __iter__(self):
        return (p for p in self._photos)

    def index(self, photo):
        return self._photos.index(photo)

    def load_photos(self):
        photo_paths = FileSystemScanner.find_supported_files(self.root_path, filter_function=self._photo_delegate.file_is_supported)
        return [self._photo_delegate(path) for path in photo_paths]

    def remove(self, photo):
        self._photos.remove(photo)


class PhotoCollectionSettings:
    FILENAME = "collection.config"

    # SettingsDescriptor descriptors automatically create, populate and update
    # a variable '_settings' which keeps all descriptor attributes in a dictionary
    # with serialized values
    _settings = dict() # initialized here just to make clear that this variable exists
    tagged_photos_root_folder = SettingsDescriptor("", lambda value: re.search(r"(\w*)", value).group(1))  # SettingsDescriptor will catch if re.search returns None
    deleted_photos_folder = SettingsDescriptor("deleted", lambda value: re.search(r"(\w*)", value).group(1))  # SettingsDescriptor will catch if re.search returns None

    def __init__(self, version, path):
        assert os.path.exists(path) and os.path.isdir(path)
        self.path = os.path.join(os.path.abspath(path), PhotoCollectionSettings.FILENAME)
        self.version = version

    @classmethod
    def load(cls, path):
        settings_path = os.path.join(path, PhotoCollectionSettings.FILENAME)
        settings = None
        # match something in the form 'key=value' or 'key=" value with spaces "'
        key_value_pair_re = re.compile(r"^\s*(\S+)\s*=\s*((?:\".*\")|(?:\S+))\s*$")
        with open(settings_path, "r") as file_:
            first_line = file_.readline()
            version_match = key_value_pair_re.match(first_line)
            assert version_match is not None
            key, version = version_match.groups()
            version = version.strip("\"")
            assert key == "version"
            assert version == "1.0"
            settings = cls(version, path)

            for line in file_.readlines():
                line_match = key_value_pair_re.match(line)
                if line_match is None:
                    continue
                key, value = line_match.groups()
                value = value.strip("\"")
                if hasattr(settings, key):
                    setattr(settings, key, value)
                else:
                    logger.warning(f"Setting {key}='{value}' is discarded by {cls.__name__}")
        settings.assert_settings()
        return settings

    def store(self):
        self.assert_settings()
        any_whitespace_match_re = re.compile(r"^.*\s.*$")
        with open(self.path, "w") as file_:
            file_.write(f"version={self.version}")
            for key in sorted(self._settings.keys()):
                file_.write(os.linesep)
                value = self._settings[key]
                if any_whitespace_match_re.match(value) or len(value) == 0:
                    value = f"\"{value}\""
                file_.write(f"{key} = {value}")

    def assert_settings(self):
        assert self.tagged_photos_root_folder != self.deleted_photos_folder

    @staticmethod
    def _find_settings_files(path):
        path_abs = os.path.abspath(os.path.normpath(path))
        assert os.path.isdir(path_abs)

        # This function searches for PhotoCollectionSettings.
        # The search starts from the directory closest to root and looks into every directory
        # in the path. If no settings are found 'above' the path (in the parent directories),
        # search continues into the children directories in the order os.walk traverses the
        # filesystem.
        settings_files = list()

        # search in parent directories
        search_paths = [path_abs]
        while True:
            path_part, folder = os.path.split(search_paths[0])
            if folder == "":
                break
            # the path closer to root gets inserted to the beginning of the list
            search_paths.insert(0, path_part)

        for p in search_paths:
            collection_settings_path = os.path.join(p, PhotoCollectionSettings.FILENAME)
            if os.path.exists(collection_settings_path):
                settings_files.append(collection_settings_path)

        # search in child directories
        for root, _, _ in os.walk(path):
            if root == search_paths[-1]:
                # avoid duplicate search results in the 'path' itself
                continue
            collection_settings_path = os.path.join(root, PhotoCollectionSettings.FILENAME)
            if os.path.exists(collection_settings_path):
                settings_files.append(collection_settings_path)
        return settings_files

    @staticmethod
    def find_settings_file(path):
        settings_files = PhotoCollectionSettings._find_settings_files(path)
        num_settings_files_found = len(settings_files)
        assert num_settings_files_found <= 1
        if num_settings_files_found == 0:
            return None
        return settings_files[0]


class PhotoCollectionMetaStorageManager:
    def __init__(self, collection_root):
        self.collection_root = collection_root
        self.collection_root_meta_storage = MetaStorage(self.collection_root)

    @staticmethod
    def _recursive_delete_empty_directories(path):
        path = os.path.abspath(path)
        assert os.path.isdir(path)
        if len(os.listdir(path)) == 0:
            parent_path = os.path.dirname(path)
            os.rmdir(path)
            PhotoCollectionMetaStorageManager._recursive_delete_empty_directories(parent_path)

    @staticmethod
    def migrate_meta_data(hash, old_meta_path, new_meta_path):
        if new_meta_path == old_meta_path:
            return

        old_photo_meta_storage = MetaStorage(old_meta_path)
        if hash not in old_photo_meta_storage._data:
            return

        new_photo_meta_storage = MetaStorage(new_meta_path)
        if hash not in new_photo_meta_storage._data or old_photo_meta_storage._data[hash].update_timestamp > new_photo_meta_storage._data[hash].update_timestamp:
            filename = old_photo_meta_storage._data[hash].filename
            meta = old_photo_meta_storage._data[hash].meta
            new_photo_meta_storage.set_meta(hash, filename, meta=meta)
        old_photo_meta_storage.delete_meta_storage_object(hash)

    def find_all_directories_with_meta_files(self, root_path=None):
        if root_path is None:
            root_path = self.collection_root
        assert os.path.exists(root_path)
        return [root for root, _, files in os.walk(root_path) if any(f.endswith(MetaStorage._meta_extension) for f in files)]

    def _cleanup_meta_data_and_directories(self, photo_index):
        meta_directories = self.find_all_directories_with_meta_files(photo_index.root_path)
        for meta_directory in meta_directories:
            meta_storage = MetaStorage(meta_directory)
            meta_storage_hashes = set(meta_storage._data.keys())
            photo_hashes_present = set(photo.get_hash() for photo in photo_index if os.path.dirname(photo.get_path()) == meta_directory)
            orphaned_meta_storage_hashes = meta_storage_hashes.difference(photo_hashes_present)
            for hash in orphaned_meta_storage_hashes:
                self.migrate_meta_data(hash, meta_directory, self.collection_root)
            self._recursive_delete_empty_directories(meta_directory)

    def _migrate_collection_root_meta_data(self, photo_index):
        for photo in photo_index:
            hash = photo.get_hash()
            if hash in self.collection_root_meta_storage._data:
                self.migrate_meta_data(hash, self.collection_root, os.path.dirname(photo.get_path()))

    def cleanup_and_migrate_meta_data_and_directories(self, photo_index):
        self._cleanup_meta_data_and_directories(photo_index)
        self._migrate_collection_root_meta_data(photo_index)


class PhotoCollectionManager:
    VERSION = "1.0"

    class CollectionError(Exception):
        pass

    def __init__(self, photo_delegate):
        self.photo_delegate = self.hook_and_connect_photo_delegate(photo_delegate)
        self.photo_index = PhotoIndex(self.photo_delegate)
        self.collection_root = None
        self.meta_storage_manager = None

    def set_collection_root(self, path):
        path_abs = os.path.abspath(path)
        collection_root = self.find_collection(path_abs)
        assert collection_root is not None
        self.collection_root = collection_root
        self.collection_settings = PhotoCollectionSettings.load(self.collection_root)
        self.meta_storage_manager = PhotoCollectionMetaStorageManager(self.collection_root)

    @staticmethod
    def create(path):
        path_abs = os.path.abspath(path)
        assert PhotoCollectionManager.find_collection(path_abs) is None
        collection_settings = PhotoCollectionSettings(PhotoCollectionManager.VERSION, path_abs)
        collection_settings.store()

    @staticmethod
    def find_collection(path):
        path_abs = os.path.abspath(path)
        try:
            settings_path = PhotoCollectionSettings.find_settings_file(path_abs)
        except AssertionError:
            settings_paths = PhotoCollectionSettings._find_settings_files(path_abs)
            raise PhotoCollectionManager.CollectionError(
                f"Multiple collection settings found: {settings_paths}.",
            )
        if settings_path is None:
            return None
        collection_root = os.path.dirname(settings_path)
        return collection_root

    def open_photos(self, path, sort_key=None):
        path_abs = os.path.abspath(path)
        assert os.path.isdir(path_abs)
        self.photo_index.load(path_abs, sort_key) 

    def delete_photo(self, photo):
        assert self.collection_root is not None
        deleted_photos_folder = self.collection_settings.deleted_photos_folder
        assert isinstance(deleted_photos_folder, str)
        if deleted_photos_folder is None or deleted_photos_folder == "":
            os.remove(photo.get_path())
        else:
            deleted_photo_path = os.path.join(
                self.collection_root,
                deleted_photos_folder,
                photo.get_filename(),
            )
            photo.set_path(deleted_photo_path)
        if photo in self.photo_index:
            self.photo_index.remove(photo)

    @staticmethod
    def delete_duplicate_photos(photo_index):
        photo_index_root_path = photo_index.root_path
        photo_index_photo_delegate = photo_index._photo_delegate
        all_photo_hashes = set(photo.get_hash() for photo in photo_index)
        to_delete = list()
        for photo in photo_index:
            photo_hash = photo.get_hash()
            if photo_hash in all_photo_hashes:
                all_photo_hashes.discard(photo_hash)
                continue
            to_delete.append(photo)
        for photo in to_delete:
            os.remove(photo.get_path())

        if len(to_delete) == 0:
            return photo_index

        new_photo_index = PhotoIndex(photo_index_photo_delegate)
        new_photo_index.load(photo_index_root_path)
        return new_photo_index

    def autosort_photos(self, photo_index):
        # Sorts all photos from photo_index into collection.
        # - photo comes from within collection:
        #   - with main tag meta data:
        #     -> move to '<collection>/<tagged_photos_root_folder>/<main_tag_photo_path>'
        #   - without main tag meta data:
        #     -> don't move photo
        # - photo comes from outside collection:
        #   - with main tag meta data:
        #     -> move to '<collection>/<tagged_photos_root_folder>/<main_tag_photo_path>'
        #   - without main tag meta data:
        #     -> move to '<collection>/<date-now>/<photo-path-within-index>'
        assert photo_index.root_path is not None
        assert self.collection_root is not None
        photo_index_root_path = photo_index.root_path
        timestamped_collection_root_path = os.path.join(
            self.collection_root,
            datetime.datetime.now().strftime("added-at-%Y-%m-%d_%H-%M-%S"),
        )

        def get_photo_index_subpath(path):
            assert path.startswith(photo_index_root_path)
            path_remainder, subpath = os.path.split(path)
            if path_remainder == photo_index_root_path:
                return subpath
            return os.path.join(get_photo_index_subpath(path_remainder), subpath)

        for photo in photo_index:
            # try to move photo based on main-tag information ...
            main_tag_photo_path = self._get_main_tag_photo_path(photo)
            if main_tag_photo_path is not None:
                photo.set_path(main_tag_photo_path)
                continue
            # ... if no main_tag_photo_path can be determined, check if photo is stored
            # within collection ...
            if photo.get_path().startswith(self.collection_root):
                continue
            # ... if not, move photo into collection.
            collection_photo_path = os.path.join(
                timestamped_collection_root_path,
                get_photo_index_subpath(photo.get_path()),
            )
            photo.set_path(collection_photo_path)

    def cleanup_and_integrate(self):
        assert self.photo_index.root_path is not None and self.meta_storage_manager is not None
        self.photo_index = self.delete_duplicate_photos(self.photo_index)
        self.meta_storage_manager.cleanup_and_migrate_meta_data_and_directories(self.photo_index)
        self.autosort_photos(self.photo_index)

    def get_list_of_all_tags(self):
        assert self.meta_storage_manager is not None
        meta_file_directories = self.meta_storage_manager.find_all_directories_with_meta_files()
        all_tags = set()
        for meta_file_directory in meta_file_directories:
            meta_storage = MetaStorage(meta_file_directory)
            hashes = MetaStorage(meta_file_directory).get_hashes()
            for hash in hashes:
                all_tags.update(meta_storage.get_meta(hash).get("tags", []))
        return sorted(all_tags)

    @staticmethod
    def _get_next_non_existing_path(path):
        new_path = path
        i = 0
        root, ext = os.path.splitext(path)
        while os.path.exists(new_path):
            root_base_match = re.match(r"^(.*?)(?:--\d+)?$", root)
            assert root_base_match is not None, "re.match failed for r\"^(.*?)(?:--\\d+)?$\" which is unexpected."
            root_base = root_base_match.group(1)
            i += 1
            new_path = f"{root_base}--{i}{ext}"
        assert not os.path.exists(new_path)
        return new_path

    def _get_main_tag_photo_path(self, photo):
        assert self.collection_root is not None
        assert isinstance(self.collection_settings.tagged_photos_root_folder, str)
        main_tag = photo.get_main_tag()
        if main_tag is None:
            return None
        
        return os.path.join(
            self.collection_root,
            self.collection_settings.tagged_photos_root_folder,  # might be empty string which os.path.join can handle properly
            *main_tag.split(":"),
            photo.get_filename(),
        )

    def hook_and_connect_photo_delegate(self, photo_delegate):
        class PhotoDelegateHooked(photo_delegate):

            #override
            @staticmethod
            def file_is_supported(path):
                assert self.collection_root is not None
                deleted_photos_folder = self.collection_settings.deleted_photos_folder
                assert isinstance(deleted_photos_folder, str)
                if deleted_photos_folder == "" or deleted_photos_folder is None:
                    return photo_delegate.file_is_supported(path)

                return (
                    photo_delegate.file_is_supported(path)
                    and not path.startswith(os.path.join(
                        self.collection_root,
                        deleted_photos_folder,
                    ))
                )

            #override
            def set_main_tag(_self, tag):
                assert self.collection_root is not None
                old_tag = _self.get_main_tag()
                super().set_main_tag(tag)
                new_path = self._get_main_tag_photo_path(_self)

                if new_path is None and old_tag is not None:
                    # The file was autosorted but should be moved into collection root
                    new_path = os.path.join(self.collection_root, _self.get_filename())

                if new_path is not None and _self.get_path() is not new_path:
                    # The file is in the wrong location
                    _self.set_path(new_path)

            #override
            def set_path(_self, path):
                assert self.meta_storage_manager is not None
                old_path = _self.get_path()
                if old_path == path:
                    return
                old_directory = os.path.dirname(old_path)
                new_directory = os.path.dirname(path)
                photo_hash = _self.get_hash()

                os.makedirs(new_directory, exist_ok=True)
                new_path = self._get_next_non_existing_path(path)
                _self._path = new_path
                shutil.move(old_path, new_path)
                self.meta_storage_manager.migrate_meta_data(photo_hash, old_directory, new_directory)
                self.meta_storage_manager._recursive_delete_empty_directories(old_directory)
        return PhotoDelegateHooked



# SPDX-FileCopyrightText: 2023 Anthony Zimmermann
#
# SPDX-License-Identifier: GPL-2.0-only

import threading
import time


class DelayedQueryManager:
    # This class basically allows to schedule computational intensive
    # tasks delayed such that they can still be cancelled within a short
    # time frame. This functionality is useful e.g., for a
    # PhotoViewer implementation where scrolling from top to bottom
    # calls a function that queries all images the way from top to
    # bottom. The DelayedQueryManger helps by delaying every query just
    # a very little bit such that it is not too late to cancel the actual
    # computationally intensive task before it starts. This way a lot
    # of scrolling would result in the creation of many delayed tasks
    # most of which get cancelled before they actually consume
    # CPU resources.
    QUERY_DELAY = 0.1 # in s delay before a query is really executed
    def __init__(self):
        self.pending_queries = list()
        self.pending_queries_lock = threading.Lock()
        self.max_queries = None

    def set_max_queries(self, value):
        self.max_queries = value

    class DelayedQuery:
        def __init__(self, query):
            self.query = query
            self.timer_thread = None
            self.stop_event = threading.Event()

        def start(self):
            def wrapped():
                time.sleep(DelayedQueryManager.QUERY_DELAY)
                if self.stop_event.is_set():
                    return
                self.query()
            self.timer_thread = threading.Thread(target=wrapped)
            self.timer_thread.start()

        def stop(self):
            if self.timer_thread is not None:
                self.stop_event.set()
            self.timer_thread = None

    def start_query_if_none_pending(self, query_key, query):
        with self.pending_queries_lock:
            matching_pending_queries = [elem for elem in self.pending_queries if elem[0] == query_key]
            if len(matching_pending_queries) == 0:
                if self.max_queries is not None:
                    while len(self.pending_queries) > self.max_queries:
                        _, delayed_query = self.pending_queries.pop(0)
                        delayed_query.stop()
                    delayed_query = self.DelayedQuery(query)
                    self.pending_queries.append((query_key, delayed_query))
                    delayed_query.start()
            else:
                # reprioritize existing queries by moving them to the end of the list
                # such that they do not get deleted next
                for matching_pending_query in matching_pending_queries:
                    self.pending_queries.remove(matching_pending_query)
                    self.pending_queries.append(matching_pending_query)

    def stop_query_if_pending(self, query_key):
        with self.pending_queries_lock:
            for matching_pending_query in (elem for elem in self.pending_queries if elem[0] == query_key):
                self.pending_queries.remove(matching_pending_query)
                _, query = matching_pending_query
                query.stop()



# SPDX-FileCopyrightText: 2023 Anthony Zimmermann
#
# SPDX-License-Identifier: GPL-2.0-only

import multiprocessing
import multiprocessing.pool
import threading
import time
import enum

from photoassistant.utils.descriptors import RDescriptor
from photoassistant.utils.descriptors import RWDescriptor
from photoassistant.utils.processingutils import AsyncExecutePolicy


class AsyncSharedCacheQueryPolicy:
    NONE = object()
    query_owner = RDescriptor()
    function = RDescriptor()
    callback = RDescriptor()
    result_placeholder = RWDescriptor()

    def __init__(self, query_owner, function, callback=None):
        self._query_owner = query_owner
        self._function = function
        self._callback = callback
        self._result_placeholder = AsyncSharedCacheQueryPolicy.NONE

    def get_query_owner(self):
        return self._query_owner

    def get_function(self):
        return self._function

    def get_callback(self):
        return self._callback

    def get_result_placeholder(self):
        return self._result_placeholder

    def set_result_placeholder(self, value):
        self._result_placeholder = value

    def delete_result_placeholder(self):
        self._result_placeholder = AsyncSharedCacheQueryPolicy.NONE


class CacheObject:
    data = RWDescriptor()
    last_access = RDescriptor()
    touched = RDescriptor()

    def __init__(self):
        self._data = None
        self._touched = threading.Event()
        self._ready = threading.Event()
        self._last_access = time.time()
        self.cache_object_lock = threading.Lock()

    def get_data(self):
        with self.cache_object_lock:
            self._touched.set()
            self._last_access = time.time()
            return self._data

    def set_data(self, data):
        with self.cache_object_lock:
            self._touched.set()
            self._last_access = time.time()
            self._data = data
            self._ready.set()

    def get_last_access(self):
        with self.cache_object_lock:
            return self._last_access

    def get_touched(self):
        return self._touched.is_set()

    def wait(self):
        return self._ready.wait()

    def is_ready(self):
        return self._ready.is_set()

    def touch(self):
        with self.cache_object_lock:
            self._touched.set()
            self._last_access = time.time()


class CacheState(enum.IntEnum):
    DATA_READY = 0
    DATA_MISSING = 1
    DATA_WAITING = 2


class SizeRestrictedCache:
    DELETE_THREAD_SLEEP_1 = 0.01
    DELETE_THREAD_SLEEP_2 = 0.5

    def __init__(self, slots):
        self.slots = slots
        self._cache = dict()
        self._cache_lock = threading.RLock()

        self._delete_old_cache_objects_worker_thread_lock = threading.Lock()
        self._delete_old_cache_objects_worker_thread = None

    def __len__(self):
        with self._cache_lock:
            return sum(sum(len(owner_function_cache) for owner_function_cache in owner_cache.values()) for owner_cache in self._cache.values())

    def get_cache_state(self, cache_owner, cache_function, key):
        with self._cache_lock:
            owner_cache = self._cache.get(cache_owner, dict())
            owner_function_cache = owner_cache.get(cache_function, dict())
            cache_object = owner_function_cache.get(key, None)
            if cache_object is None:
                return CacheState.DATA_MISSING
            if cache_object.is_ready():
                return CacheState.DATA_READY
            else:
                return CacheState.DATA_WAITING

    def query_cache_object(self, cache_owner, cache_function, key):
        with self._cache_lock:
            owner_cache = self._cache.get(cache_owner, dict())
            owner_function_cache = owner_cache.get(cache_function, dict())
            cache_object = owner_function_cache.get(key, CacheObject())
            owner_function_cache[key] = cache_object
            owner_cache[cache_function] = owner_function_cache
            self._cache[cache_owner] = owner_cache

            self._delete_old_cache_objects_in_separate_thread()

        return cache_object

    def _delete_old_cache_objects_loop(self):
        # The reference 'self._delete_old_cache_objects_worker_thread' will get deleted when this loop exits.
        # Other calls of '_delete_old_cache_objects_in_separate_thread' would create a new thread if the reference is None.
        # To prevent spawning dozens of short-lived threads, this thread is kept alive a little longer than necessary.
        # (*sleep_1) DELETE_THREAD_SLEEP_1:
        #     One short sleep is within the loop. The sleep is skipped if there are too many slots still to be deleted.
        # (*sleep_2) DELETE_THREAD_SLEEP_2:
        #     A long sleep keeps the thread awake just before it would stop and checks again if there are entries
        #     to be deleted. This trades off keeping a thread alive with a longer time vs. spawning many new threads
        #     when there are lots of cache entries added in a burst.
        while True:
            # release the lock after every access such that this thread does not block
            # access to the cache for long
            occupied_slots = 0
            with self._cache_lock:
                occupied_slots = len(self)

            if occupied_slots <= self.slots:
                time.sleep(self.DELETE_THREAD_SLEEP_2)  # (*sleep_2)
                with self._cache_lock:
                    occupied_slots = len(self)

            if occupied_slots <= self.slots:
                # No cache items have been added during (*sleep_2)
                # -> Thread can terminate. A new thread will potentially get spawned with the next cache access.
                break

            with self._cache_lock:
                delete_candidates = []
                for owner, owner_cache in self._cache.items():
                    if len(owner_cache) == 0:
                        continue
                    for cache_function, owner_function_cache in owner_cache.items():
                        if len(owner_function_cache) == 0:
                            continue
                        least_recently_used_key, cache_obj = min(owner_function_cache.items(), key=lambda item: item[1].last_access)
                        delete_candidates.append(((owner, cache_function, least_recently_used_key), cache_obj))
                (delete_owner, delete_cache_function, delete_key), _ = min(delete_candidates, key=lambda item: item[1].last_access)
                del self._cache[delete_owner][delete_cache_function][delete_key]
                if len(self._cache[delete_owner][delete_cache_function]) == 0:
                    del self._cache[delete_owner][delete_cache_function]
                    if len(self._cache[delete_owner]) == 0:
                        del self._cache[delete_owner]

                if len(self) > (self.slots + 10):
                    # Skip (*sleep_1) as there is a lot work still to do.
                    continue

            time.sleep(self.DELETE_THREAD_SLEEP_1)  # (*sleep_1)
        # remove reference such that a new thread can be started
        with self._delete_old_cache_objects_worker_thread_lock:
            self._delete_old_cache_objects_worker_thread = None

    def _delete_old_cache_objects_in_separate_thread(self):
        with self._cache_lock:
            if len(self) <= self.slots:
                return
        with self._delete_old_cache_objects_worker_thread_lock:
            if self._delete_old_cache_objects_worker_thread is not None:
                return
            self._delete_old_cache_objects_worker_thread = threading.Thread(target=self._delete_old_cache_objects_loop)
        self._delete_old_cache_objects_worker_thread.start()

    def delete(self, cache_owner):
        with self._cache_lock:
            if cache_owner not in self._cache:
                return
            del self._cache[cache_owner]


class AsyncSharedCache:
    DEFAULT_THREADPOOL_WORKERS = multiprocessing.pool.ThreadPool(multiprocessing.cpu_count() - 1)
    DEFAULT_THREADPOOL_WAITERS = multiprocessing.pool.ThreadPool(10)
    slots = RWDescriptor()

    def __init__(self, slots, worker_pool=None, waiter_pool=None):
        self._cache = SizeRestrictedCache(slots)
        self.worker_pool = worker_pool if worker_pool is not None else AsyncSharedCache.DEFAULT_THREADPOOL_WORKERS
        self.waiter_pool = waiter_pool if waiter_pool is not None else AsyncSharedCache.DEFAULT_THREADPOOL_WAITERS

    def get_slots(self):
        return self._cache.slots

    def set_slots(self, value):
        self._cache.slots = value

    def get_cache_state(self, async_shared_cache_query_policy, *args, **kwargs):
        key = (args, tuple(sorted(kwargs.items())))
        return self._cache.get_cache_state(
            async_shared_cache_query_policy.query_owner,
            async_shared_cache_query_policy.function,
            key,
        )

    def query(self, async_shared_cache_query_policy, *args, **kwargs):
        key = (args, tuple(sorted(kwargs.items())))
        cache_object = self._cache.query_cache_object(
            async_shared_cache_query_policy.query_owner,
            async_shared_cache_query_policy.function,
            key,
        )

        cache_hit = cache_object.touched # cache already has an entry for this query, the entry might be ready or not at this point
        cache_object.touch() # touch as early as possible such that other calls will not try to execute the actual query too
        # data is already in cache and ready to be returned
        # -> asynchroneously execute callback, return data
        if cache_object.is_ready():
            if async_shared_cache_query_policy.callback is not None:
                # still execute the callback function as requested
                AsyncExecutePolicy(async_shared_cache_query_policy.callback)(self.worker_pool, cache_object.data)
            return cache_object.data
        # data is not already in cache
        # no placeholder can be returned
        # -> valid data must be returned immediately: synchroneously fetch data and put data into cache, asynchroneously execute callback, return data
        elif async_shared_cache_query_policy.result_placeholder == AsyncSharedCacheQueryPolicy.NONE:
            if not cache_hit:
                cache_object.set_data(async_shared_cache_query_policy.function(*args, **kwargs))
            cache_object.wait()
            if async_shared_cache_query_policy.callback is not None:
                AsyncExecutePolicy(async_shared_cache_query_policy.callback)(self.waiter_pool, cache_object.data)
            return cache_object.data
        # data is not already in cache
        # placeholder can be returned
        # -> cache entry is freshly created and must be filled with data: asynchroneously fetch data, put data into cache and execute callback, return placeholder
        elif not cache_hit: # cache is missing data
            def callback_wrapper(result):
                cache_object.set_data(result)
                if async_shared_cache_query_policy.callback is not None:
                    async_shared_cache_query_policy.callback(result)
            AsyncExecutePolicy(
                async_shared_cache_query_policy.function,
                callback_wrapper,
            )(self.worker_pool, *args, **kwargs)
            return async_shared_cache_query_policy.result_placeholder
        # data is not already in cache
        # placeholder can be returned
        # cache entry is not fresh
        # -> fetching of data is already in progress: asynchroneously wait for data and execute callback, return placeholder
        else:
            if async_shared_cache_query_policy.callback is not None:
                def wait_for_cache_data():
                    cache_object.wait()
                    return cache_object.data
                AsyncExecutePolicy(
                    wait_for_cache_data,
                    async_shared_cache_query_policy.callback,
                )(self.waiter_pool)
            return async_shared_cache_query_policy.result_placeholder

    def delete_cache(self, cache_owner):
        self._cache.delete(cache_owner)
